"""Remove some pipeline directories."""
from concurrent.futures import ThreadPoolExecutor
from glob import glob
from os import path
from ..utils import fs
from ..plugin import hookimpl
from ..logger import logger
from .list import get_procs, show_proc


def remove_proc(proc, nthread=1, force=False):
    """Remove process from directory"""
    if force:
        with ThreadPoolExecutor(max_workers=nthread) as executor:
            for folder in glob(path.join(proc, '*')):
                if path.isdir(folder):
                    executor.submit(fs.remove, folder)
        try:
            fs.remove(proc)
            logger.warning('  Removed!           ')
        # pylint: disable=broad-except
        except BaseException as ex:  # pragma: no cover
            #shutil.rmtree(proc)
            logger.error('  %s!           ' % ex)

    else:
        ans = input('  Remove it? [Y/n] ')
        print('\r')
        while ans not in ('', 'Y', 'y', 'N', 'n'):
            ans = input('  Remove it? [Y/n] ')
            print('\r')
        if ans in ['', 'Y', 'y']:
            remove_proc(proc, nthread, True)


def clean_procs(procs, nthread, force, wdir, one):
    """Cleanup processes"""
    logger.workdir(wdir)
    procname = None
    for proc, mtime, fail in procs:
        process_name = '.'.join(path.basename(proc).split('.')[1:3])
        if process_name != procname:
            procname = process_name
            logger.process(process_name)
            show_proc(proc, mtime, fail)
            if not one:
                remove_proc(proc, nthread, force)
        else:
            show_proc(proc, mtime, fail)
            remove_proc(proc, nthread, force)


@hookimpl
def cli_addcmd(commands):
    """Add command"""
    commands.clean = __doc__
    commands.clean.nthread = 1
    commands.clean.nthread.desc = ('Number of threads used to clean up '
                                   'the work directories.')
    commands.clean.proc = commands.list.proc
    commands.clean.ago = commands.list.ago
    commands.clean.before = commands.list.before
    commands.clean.nocheck = commands.list.nocheck
    commands.clean.force = False
    commands.clean.force.desc = 'Don`t ask when remove work directories.'
    commands.clean.error = commands.list.error
    commands.clean.wdir = commands.list.wdir
    commands.clean.one = False
    commands.clean.one.desc = 'Just keep one process under a process group.'


@hookimpl
def cli_execcmd(command, opts):
    """Run the command"""
    if command == 'clean':
        clean_procs(get_procs(opts), opts.nthread, opts.force, opts.wdir,
                    opts.one)



"""
Channel for pyppl
"""
import sys
import functools

from os import path
from glob import glob
from liquid.stream import safe_split

# pylint: disable=invalid-name


class Channel(list):  # pylint: disable=too-many-public-methods
    """@API
    The channen class, extended from `list`
    """
    @staticmethod
    def _tuplize(atuple):
        """@API
        A private method, try to convert an element to tuple
        If it's a string, convert it to `(atuple, )`
        Else if it is iterable, convert it to `tuple(atuple)`
        Otherwise, convert it to `(atuple, )`
        Notice that string is also iterable.
        @params:
            atuple (str|list|tuple): the element to be converted
        @returns:
            (tuple): The converted element
        """
        if isinstance(atuple, (str, list)):
            return (atuple, )

        try:
            iter(atuple)
        except TypeError:
            return (atuple, )
        return tuple(atuple)

    @staticmethod
    def create(alist=None):
        """@API
        Create a Channel from a list
        @params:
            alist (list|Channel): The list, default: []
        @returns:
            (Channel): The Channel created from the list
        """
        if alist is None:
            alist = []

        if not isinstance(alist, list):
            alist = [alist]

        ret = Channel()
        length = 0
        for ele in alist:
            row = Channel._tuplize(ele)
            if length == 0:
                length = len(row)
            if length != len(row):
                raise ValueError(
                    'Inconsistent width of row (%s) with previous row (%s)' %
                    (len(row), length))
            ret.append(row)
        return ret

    @staticmethod
    def nones(length, width):
        """@API
        Create a channel with `None`s
        @params:
            length (int): The length of the channel
            width (int):  The width of the channel
        @returns:
            (Channel): The created channel
        """
        ret = Channel()
        row = (None, ) * width
        for _ in range(length):
            ret.append(row[:])
        return ret

    @staticmethod
    def fromChannels(*args):
        """@API
        Create a Channel from Channels
        @params:
            *args (any): The Channels or anything can be created as a `Channel`
        @returns:
            (Channel): The Channel merged from other Channels
        """
        ret = Channel.create()
        return ret.insert(None, *args)

    @staticmethod
    def fromPattern(pattern, ftype='any', sortby='name', reverse=False):
        """@API
        Create a Channel from a path pattern
        @params:
            pattern (str): the pattern with wild cards
            ftype (str): the type of the files/dirs to include
              - 'dir', 'file', 'link' or 'any' (default)
            sortby (str):  how the list is sorted
              - 'name' (default), 'mtime', 'size'
            reverse (bool): reverse sort. Default: `False`
        @returns:
            (Channel): The Channel created from the path
        """
        if sortby == 'name':
            key = str
        elif sortby == 'mtime':
            key = path.getmtime
        elif sortby == 'size':
            key = path.getsize

        filt = lambda afile: True
        if ftype == 'link':
            filt = path.islink
        elif ftype == 'dir':
            filt = lambda afile: path.isdir(afile) and not path.islink(afile)
        elif ftype == 'file':
            filt = lambda afile: path.isfile(afile) and not path.islink(afile)
        files = [afile for afile in glob(str(pattern)) if filt(afile)]
        return Channel.create(sorted(files, key=key, reverse=reverse))

    @staticmethod
    def fromPairs(pattern):
        """@API
        Create a width = 2 Channel from a pattern
        @params:
            pattern (str): the pattern
        @returns:
            (Channel): The Channel create from every 2 files match the pattern
        """
        ret = sorted(glob(pattern))
        chan = Channel.create()
        for i in range(0, len(ret), 2):
            chan.append((ret[i], ret[i + 1]))
        return chan

    @staticmethod
    def fromFile(filename, header=False, skip=0, delimit="\t"):
        """@API
        Create Channel from the file content
        It's like a matrix file, each row is a row for a Channel.
        And each column is a column for a Channel.
        @params:
            filename (file): the file
            header (bool):  Whether the file contains header.
                If True, will attach the header
                - So you can use `channel.<header>` to fetch the column
            skip (int): first lines to skip, default: `0`
            delimit (str): the delimit for columns, default: `\t`
        @returns:
            (Channel): A Channel created from the file
        """
        ret = Channel.create()
        i = -1
        headers = []
        with open(filename) as fhandler:
            for line in fhandler:
                i += 1
                if i < skip:
                    continue
                line = line.rstrip("\n")
                if not line:
                    continue
                if header:
                    headers = line.split(delimit)
                    header = False
                else:
                    ret.append(tuple(line.split(delimit)))
        if headers:
            lenheaders = len(headers)
            channelwid = ret.width()
            if lenheaders == channelwid - 1:
                headers.insert(0, 'RowNames')
            elif lenheaders != channelwid:
                raise ValueError(
                    'Number of headers and columns doesn\'t match: %s, %s.' %
                    (lenheaders, channelwid))
            ret.attach(*headers)

        return ret

    @staticmethod
    def fromArgv():
        """@API
        Create a Channel from `sys.argv[1:]`
        "python test.py a b c" creates a width=1 Channel
        "python test.py a,1 b,2 c,3" creates a width=2 Channel
        @returns:
            (Channel): The Channel created from the command line arguments
        """
        ret = Channel.create()
        args = sys.argv[1:]
        alen = len(args)
        if alen == 0:
            return ret

        width = None
        for arg in args:
            items = tuple(safe_split(arg, ','))
            if width is not None and width != len(items):
                raise ValueError(
                    'Width %s (%s) is not consistent with previous %s' %
                    (len(items), arg, width))
            width = len(items)
            ret.append(items)
        return ret

    @staticmethod
    def fromParams(*pnames):
        """@API
        Create a Channel from params
        @params:
            *pnames (str): The names of the option
        @returns:
            (Channel): The Channel created from `pyparam`.
        """
        from pyparam import params
        ret = Channel.create()
        width = None
        for pname in pnames:
            param = getattr(params, pname)
            data = param.value
            if not param.type.startswith('list') or not isinstance(data, list):
                data = [param.value]
            if width is not None and width != len(data):
                raise ValueError(
                    'Width %s (%s) is not consistent with previous %s' %
                    (len(data), data, width))
            width = len(data)
            ret = ret.cbind(data)
        if ret:
            ret.attach(*pnames)
        return ret

    # pylint: disable=too-many-arguments
    def expand(self,
               col=0,
               pattern="*",
               ftype='any',
               sortby='name',
               reverse=False):
        """@API
        expand the Channel according to the files in <col>,
        other cols will keep the same
        `[(dir1/dir2, 1)].expand (0, "*")` will expand to
        `[(dir1/dir2/file1, 1), (dir1/dir2/file2, 1), ...]`
        length: 1 -> N
        width:  M -> M
        @params:
            col (int): the index of the column used to expand
            pattern (str): use a pattern to filter the files/dirs, default: `*`
            ftype (str): the type of the files/dirs to include
              - 'dir', 'file', 'link' or 'any' (default)
            sortby (str):  how the list is sorted
              - 'name' (default), 'mtime', 'size'
            reverse (bool): reverse sort. Default: False
        @returns:
            (Channel): The expanded Channel
        """
        ret = Channel.create()
        if not self:
            return ret

        for row in self:
            row = list(row)
            folder = row[col]
            files = Channel.fromPattern(path.join(folder, pattern),
                                        ftype=ftype,
                                        sortby=sortby,
                                        reverse=reverse).flatten()
            for afile in files:
                row[col] = afile
                ret.append(tuple(row))
        return ret

    def collapse(self, col=0):
        """@API
        Do the reverse of expand
        length: N -> 1
        width:  M -> M
        @params:
            col (int):     the index of the column used to collapse
        @returns:
            (Channel): The collapsed Channel
        """
        if not self:
            raise ValueError('Cannot collapse an empty Channel.')

        paths = self.colAt(col).flatten()
        compx = path.dirname(path.commonprefix(paths))
        row = list(self[0])
        row[col] = compx
        return Channel.create(tuple(row))

    def copy(self):
        """@API
        Copy a Channel using `copy.copy`
        @returns:
            (Channel): The copied Channel
        """
        return self.slice(0)

    def width(self):
        """@API
        Get the width of a Channel
        @returns:
            (int): The width of the Channel
        """
        if not self:
            return 0
        return len(self[0]) if isinstance(self[0], tuple) else 1

    def length(self):
        """@API
        Get the length of a Channel
        It's just an alias of `len(chan)`
        @returns:
            (int): The length of the Channel
        """
        return len(self)

    def map(self, func):
        """@API
        Alias of python builtin `map`
        @params:
            func (callable): the function
        @returns:
            (Channel): The transformed Channel
        """
        return Channel.create([func(x) for x in self])

    def mapCol(self, func, col=0):
        """@API
        Map for a column
        @params:
            func (callable): the function
            col (int): the index of the column. Default: `0`
        @returns:
            (Channel): The transformed Channel
        """
        return Channel.create(
            [s[:col] + (func(s[col]), ) + s[(col + 1):] for s in self])

    def filter(self, func=None):
        """@API
        Alias of python builtin `filter`
        @params:
            func (callable): the function. Default: `None`
        @returns:
            (Channel): The filtered Channel
        """
        func = func or bool
        return Channel.create([x for x in self if func(x)])

    def filterCol(self, func=None, col=0):
        """@API
        Just filter on the specific column
        @params:
            func (callable): the function
            col (int): the column to filter
        @returns:
            (Channel): The filtered Channel
        """
        func = func or bool
        return Channel.create([s for s in self if func(s[col])])

    def reduce(self, func):
        """@API
        Alias of python builtin `reduce`
        @params:
            func (callable): the function
        @returns:
            (Channel): The reduced value
        """
        return functools.reduce(func, self)

    def reduceCol(self, func, col=0):
        """@API
        Reduce a column
        @params:
            func (callable): the function
            col (int): the column to reduce
        @returns:
            (Channel): The reduced value
        """
        return functools.reduce(func, [s[col] for s in self])

    def rbind(self, *rows):
        """@API
        The multiple-argument versoin of `rbind`
        @params:
            *rows (any): the rows to be bound to Channel
        @returns:
            (Channel): The combined Channel
        """
        ret = self.copy()
        for row in rows:
            if not row:
                continue
            if not isinstance(row, Channel):
                row = Channel.create(row)
            if row.length() == 0 or row.width() == 0:
                continue

            if ret.length() == 0:
                ret = row
            elif row.width() == 1:
                ret.extend([rele * ret.width() for rele in row])
            elif ret.width() == 1:
                for i in range(ret.length()):
                    ret[i] *= row.width()
                ret.extend(row)
            elif row.width() != ret.width():
                raise ValueError(
                    'Unable to rbind unequal width channels (%s, %s)' %
                    (ret.width(), row.width()))
            else:
                ret.extend(row)
        return ret

    def insert(self, cidx, *cols):
        """@API
        Insert columns to a channel
        @params:
            cidx (int): Insert into which index of column?
            *cols (any): the columns to be bound to Channel
        @returns:
            (Channel): The combined Channel
        """
        ret = self.copy()
        #cols = [col if isinstance(col, Channel)
        # else Channel.create(col) for col in cols if col]
        # fix #29
        cols = [
            col if isinstance(col, Channel) else Channel.create(col)
            for col in cols
        ]
        cols = [col for col in cols if col.width() > 0 and col.length() > 0]
        if not cols:
            return ret

        if cidx is None:
            colsbefore = ret
            colsafter = []
        else:
            colsbefore = [s[:cidx] for s in ret]
            colsafter = [s[cidx:] for s in ret]

        maxlen = max([col.length() for col in cols])
        maxlen = max(maxlen, len(ret))
        if ret.length() == 1:
            colsbefore *= maxlen
            colsafter *= maxlen

        length = max(len(colsbefore), len(colsafter))
        for col in cols:
            # has been filtered with col.width() > 0
            #if col.width() == 0:
            #	continue
            if col.length() == 1:
                # oringal col has been changed
                #col *= maxlen
                col = col.repRow(maxlen)
            if length > 0 and length != col.length():
                raise ValueError(
                    'Cannot bind column (length: %s) to Channel (length: %s).'
                    % (col.length(), length))

            colsbefore = [colsbefore[i] + c
                          for i, c in enumerate(col)] if colsbefore else col
            length = max(len(colsbefore), len(colsafter))

        return Channel(
            [c + colsafter[i]
             for i, c in enumerate(colsbefore)] if colsafter else colsbefore)

    def cbind(self, *cols):
        """@API
        Add columns to the channel
        @params:
            *cols (any): The columns
        @returns:
            (Channel): The channel with the columns inserted.
        """
        return self.insert(None, *cols)

    def colAt(self, index):
        """@API
        Fetch one column of a Channel
        @params:
            index (int|list): which column to fetch
        @returns:
            (Channel): The Channel with that column
        """
        if not isinstance(index, list):
            index = [index]
        chs = []
        for idx in index:
            chs.append(self.slice(idx, 1))
        return Channel.fromChannels(*chs)

    def rowAt(self, index):
        """@API
        Fetch one row of a Channel
        @params:
            index (int): which row to fetch
        @returns:
            (Channel): The Channel with that row
        """
        if not isinstance(index, list):
            index = [index]
        rows = []
        for idx in index:
            rows.extend(self[idx:(idx + 1 if idx != -1 else None)])
        return Channel(rows)

    def unique(self):
        """@API
        Make the channel unique, remove duplicated rows
        Try to keep the order
        @returns:
            (Channel): The channel with unique rows.
        """
        rows = []
        for row in self:
            if row not in rows:
                rows.append(row)
        return Channel(rows)

    def slice(self, start, length=None):
        """@API
        Fetch some columns of a Channel
        @params:
            start (int):  from column to start
            length (int): how many columns to fetch,
                default: None (from start to the end)
        @returns:
            (Channel): The Channel with fetched columns
        """
        return (Channel([s[start:(start + length)] for s in self])
                if length and start != -1
                else Channel.create([s[start:] for s in self]))

    def fold(self, nfold=1):
        """@API
        Fold a Channel. Make a row to n-length chunk rows
        ```
        a1	a2	a3	a4
        b1	b2	b3	b4
        if nfold==2, fold(2) will change it to:
        a1	a2
        a3	a4
        b1	b2
        b3	b4
        ```
        @params:
            nfold (int): the size of the chunk
        @returns
            (Channel): The new Channel
        """
        if nfold <= 0 or self.width() % nfold != 0:
            raise ValueError(
                'Failed to fold, the width %s cannot be divided by %s' %
                (self.width(), nfold))

        nrows = int(self.width() / nfold)
        ret = []
        for row in self:
            for i in [x * nfold for x in range(nrows)]:
                ret.append(row[i:i + nfold])
        return Channel(ret)

    def unfold(self, nfold=2):
        """@API
        Do the reverse thing as self.fold does
        @params:
            nfold (int): How many rows to combind each time. default: 2
        @returns:
            (Channel): The unfolded Channel
        """
        if nfold <= 0 or len(self) % nfold != 0:
            raise ValueError(
                'Failed to unfold, the length %s cannot be divided by %s' %
                (len(self), nfold))

        nrows = int(len(self) / nfold)
        ret = []
        for i in range(nrows):
            row = ()
            for j in range(nfold):
                row += self[i * nfold + j]
            ret.append(row)
        return Channel.create(ret)

    def split(self, flatten=False):
        """@API
        Split a Channel to single-column Channels
        @returns:
            (list[Channel]): The list of single-column Channels
        """
        return ([self.flatten(i) for i in range(self.width())]
                if flatten
                else [self.colAt(i) for i in range(self.width())])

    def attach(self, *names, flatten=False):
        """@API
        Attach columns to names of Channel, so we can access each column by:
        `ch.col0` == ch.colAt(0)
        @params:
            *names (str): The names. Have to be as length as channel's width.
                None of them should be Channel's property name
            flatten (bool): Whether flatten the channel for
                the name being attached
        """
        mywidth = self.width()
        lnames = len(names)
        if mywidth < lnames:
            raise ValueError(
                'More names (%s) to attach than the width (%s) of the channel.'
                % (lnames, mywidth))
        for i, name in enumerate(names):
            if (hasattr(self, name)
                    and not isinstance(getattr(self, name), Channel)
                    and not isinstance(getattr(self, name), list)):
                raise AttributeError(
                    'Cannot attach column to "%s" as it is already '
                    'an attribute of Channel.' % name)
            setattr(self, name, self.flatten(i) if flatten else self.colAt(i))

    def get(self, idx=0):
        """@API
        Get the element of a flattened channel
        @params:
            idx (int): The index of the element to get. Default: 0
        @return:
            (any): The element
        """
        return self.flatten()[idx]

    def repCol(self, nrep=2):
        """@API
        Repeat column and return a new channel
        @params:
            nrep (int): how many times to repeat.
        @returns:
            (Channel): The new channel with repeated columns
        """
        cols = [self] * (nrep - 1)
        return self.cbind(*cols)

    def repRow(self, nrep=2):
        """@API
        Repeat row and return a new channel
        @params:
            nrep (int): how many times to repeat.
        @returns:
            (Channel): The new channel with repeated rows
        """
        rows = [self] * (nrep - 1)
        return self.rbind(*rows)

    def flatten(self, col=None):
        """@API
        Convert a single-column Channel to a list (remove the tuple signs)
        `[(a,), (b,)]` to `[a, b]`
        @params:
            col (int): The column to flat. None for all columns (default)
        @returns:
            (list): The list converted from the Channel.
        """
        return ([item for sublist in self for item in sublist]
                if col is None
                else [sublist[col] for sublist in self])

    def transpose(self):
        """@API
        Transpose the channel
        @returns:
            (Channel): The transposed channel.
        """
        ret = Channel.nones(length=self.width(), width=self.length())
        ret = [list(row) for row in ret]
        for i, row in enumerate(self):
            for j, val in enumerate(row):
                ret[j][i] = val
        ret = [tuple(row) for row in ret]
        return Channel(ret)

    # We will try to deprecate the camelCase functions
    t = transpose
    from_pattern = fromPattern
    from_pairs = fromPairs
    from_file = fromFile
    from_argv = fromArgv
    from_params = fromParams
    from_channels = fromChannels
    map_col = mapCol
    filter_col = filterCol
    reduce_col = reduceCol
    col_at = colAt
    row_at = rowAt
    rep_col = repCol
    rep_row = repRow



"""Check the status of a process."""
from glob import glob
from os import path
from ..plugin import hookimpl
from ..logger import logger


@hookimpl
def cli_addcmd(commands):
    """Add command"""
    commands.status = 'Check the status of a running process.'
    commands.status.ncol = 5
    commands.status.ncol.desc = 'Number of jobs to show in a row.'
    commands.status.proc.required = True
    commands.status.proc.desc = ('The process working directory. If path '
                                 'separator exists, then `-wdir` will be '
                                 'ignored.')
    commands.status.wdir = commands.list.wdir


@hookimpl
def cli_execcmd(command, opts):  # pylint: disable=too-many-locals
    """Run the command"""
    if command == 'status':
        if path.sep in opts.proc:
            procdir = opts.proc
        else:
            proc = opts.proc if opts.proc.startswith(
                'PyPPL.') else 'PyPPL.' + opts.proc
            proc = glob(path.join(opts.wdir, proc + '*'))
            if len(proc) > 1:
                logger.warning(
                    f'There are more than 1 processes named with '
                    f'"{opts.proc}", first one used.'
                )
            procdir = proc[0]
        logger.workdir(procdir)

        jobdirs = list(
            sorted(glob(path.join(procdir, '*', '')),
                   key=lambda x: int(path.basename(x[:-1]))))
        nnn = len(str(len(jobdirs)))
        counts = {
            'Unknown': 0,
            'Pending': 0,
            'Running': 0,
            'Done': 0,
            'Failed': 0,
        }
        logstr = ''
        for i, jobdir in enumerate(jobdirs):
            jobdir = path.normpath(jobdir)
            pidfile = path.join(jobdir, 'job.pid')
            outfile = path.join(jobdir, 'job.stdout')
            errfile = path.join(jobdir, 'job.stderr')
            jobid = path.basename(jobdir)
            jstat = 'Unknown'
            rc = '-'
            if not path.isfile(pidfile) or not path.isfile(
                    outfile) or not path.isfile(errfile):
                jstat = 'Pending'
            else:
                rcfile = path.join(jobdir, 'job.rc')
                if not path.isfile(rcfile):
                    jstat = 'Running'
                else:
                    with open(rcfile) as frc:
                        rc = frc.read().strip()
                    jstat = 'Done' if rc == '0' else 'Failed'
            counts[jstat] += 1

            jobstr = ('#' + jobid).rjust(nnn + 1)
            logstr += jobstr + ': ' + jstat.ljust(8) + ('[' + rc +
                                                        ']    ').rjust(8)
            if (i + 1) % 4 == 0:
                logger.info(logstr)
                logstr = ''
        if logstr:
            logger.info(logstr)

        logger.info('')
        logger.info('Total: ', end='')
        for key, count in counts.items():
            logger.info('- ' + key.ljust(8) + ': ' + str(count))
        logger.info('')



"""Utility functions for PyPPL"""
import re
from os import path, walk
from copy import deepcopy
from queue import PriorityQueue
from threading import Thread
import shutil
from transitions import Transition, Machine
from liquid.stream import safe_split
from . import _fsutil as fs


def name2filename(name):
    """@API
    Convert any name to a valid filename
    @params:
        name (str): The name to be converted
    @returns:
        (str): The converted name
    """
    name = re.sub(r'[^\w_]+', '_', name)
    name = re.sub(r'_+', '_', name)
    return name.strip('_')

def log_msg_len(procid='', # pylint: disable=too-many-arguments
                joblen=0,
                ret='with-pbar',
                fallback=50,
                maxlen=80,
                minlen=20):
    """@API
    Get the progress bar size (length) according to the terminal width
    @params:
        procid (str): The process id
        joblen (int): The number of jobs
        fallback (int): The value to return if we fail
        adjust (int): Adjust the length to get the length of message with
                      or without proc id and/or job index indicators
            - `with-pbar`: return the progress bar size with everything
            - `with-nothing`: return the length without process id nor job
                              index indicator
            - `with-proc`: return the length of message with proc id only
            - `with-job`: return the length of message with proc id and
                          job index indicator
        maxlen (int): The maximum length of returned length
        minlen (int): The minimum length of returned length
    @returns:
        (int): The progress bar size
    """
    # pylint: disable=line-too-long
    # calculate # chars other than the pbar
    # typically:
    # [05-27 17:59:58    MAIN.JOBDONE] pProcess: [3/3] [ ...pbar... ] Done: 100.0% | Running: 0
    # ---------------------------------<procid>---j-j---............--------------------------j
    # pylint: enable=line-too-long
    proclen = len(procid)
    jobidlen = len(str(joblen))

    if ret == 'with-pbar':
        nonpbar_len = (33 + # logformat
                       proclen+2 + # proc id
                       1+jobidlen+1+jobidlen+2 + # job index indicator
                       1 + # [
                       2 + # ]<space>
                       24+jobidlen) # Done ...
    elif ret == 'with-nothing':
        nonpbar_len = 33
    elif ret == 'with-proc':
        nonpbar_len = 33 + proclen+2
    elif ret == 'with-job':
        nonpbar_len = 33 + proclen+2 + 2*jobidlen+4

    total_fallback = nonpbar_len + fallback

    try:
        total_width = shutil.get_terminal_size().columns - 1
    except (OSError, AttributeError): # pragma: no cover
        total_width = total_fallback

    if not isinstance(total_width, int) or total_width <= nonpbar_len:
        total_width = total_fallback
    return max(minlen, min(maxlen, total_width - nonpbar_len))

def format_secs(seconds):
    """@API
    Format a time duration
    @params:
        `seconds`: the time duration in seconds
    @returns:
        The formated string.
        For example: "01:01:01.001" stands for 1 hour 1 min 1 sec and 1 minisec.
    """
    minute, sec = divmod(seconds, 60)
    hour, minute = divmod(minute, 60)
    return "%02d:%02d:%02d.%03.0f" % (hour, minute, sec, 1000 *
                                      (sec - int(sec)))


def filesig(filepath, dirsig=True):
    """@API
    Generate a signature for a file
    @params:
        `dirsig`: Whether expand the directory? Default: True
    @returns:
        The signature
    """
    if not filepath:
        return ['', 0]
    if not fs.exists(filepath):
        return False

    getmtime = path.getmtime
    if dirsig and fs.isdir(filepath):
        mtime = getmtime(filepath)
        for root, dirs, files in walk(filepath):
            for directory in dirs:
                mtime2 = getmtime(path.join(root, directory))
                mtime = max(mtime, mtime2)
            for filename in files:
                # links to non-existent files
                if not fs.exists(path.join(root, filename)):
                    continue
                mtime2 = getmtime(path.join(root, filename))
                mtime = max(mtime, mtime2)
    else:
        mtime = getmtime(filepath)
    return [str(filepath), int(mtime)]


def funcsig(func):
    """@API
    Get the signature of a function
    Try to get the source first, if failed, try to get its name,
    otherwise return None
    @params:
        `func`: The function
    @returns:
        The signature
    """
    if callable(func):
        try:
            from inspect import getsource
            sig = getsource(func).strip()
        except (TypeError, ValueError):  # pragma: no cover
            sig = func.__name__
    else:
        sig = 'None'
    return sig


def always_list(data, trim=True):
    """@API
    Convert a string or a list with element
    @params:
        `data`: the data to be converted
        `trim`: trim the whitespaces for each item or not. Default: True
    @examples:
        ```python
        data = ["a, b, c", "d"]
        ret  = always_list (data)
        # ret == ["a", "b", "c", "d"]
        ```
    @returns:
        The split list
    """
    if isinstance(data, str):
        return safe_split(data, ',', trim=trim)
    if isinstance(data, list):
        return sum(
            (always_list(dat, trim) if isinstance(dat, (str, list)) else [dat]
             for dat in data), [])
    raise ValueError('Expect str/list to convert to list, but got %r.' %
                     type(data).__name__)


def try_deepcopy(obj, depth=3):
    """@API
    Try do deepcopy an object. If fails, just do a shallow copy.
    @params:
        obj (any): The object
        depth (int): A flag to avoid deep recursion
    @returns:
        (any): The copied object
    """
    if depth <= 0:
        return obj
    if isinstance(obj, dict):
        # do a shallow copy first
        # we don't start with an empty dictionary, because obj may be
        # an object from a class extended from dict
        ret = obj.copy()
        for key, value in obj.items():
            ret[key] = try_deepcopy(value, depth - 1)
        return ret
    if isinstance(obj, list):
        ret = obj[:]
        for i, value in enumerate(obj):
            ret[i] = try_deepcopy(value, depth - 1)
        return ret
    try:
        return deepcopy(obj)
    except BaseException:
        return obj


def chmod_x(filepath):
    """@API
    Convert file1 to executable or add extract shebang to cmd line
    @params:
        filepath (path): The file path
    @returns:
        (list): with or without the path of the interpreter as the first element
        and the script file as the last element
    """
    from stat import S_IEXEC
    from os import chmod, stat
    filepath = str(filepath)
    if not fs.isfile(filepath):
        raise OSError('Unable to make {} as executable'.format(filepath))
    # in case it's a Path-like object
    ret = [filepath]
    try:
        chmod(filepath, stat(filepath).st_mode | S_IEXEC)
    except (OSError, PermissionError):
        shebang = None
        with open(filepath) as fsb:
            try:
                shebang = fsb.readline().strip()
            except (OSError, PermissionError, UnicodeDecodeError):
                # may raise UnicodeDecodeError for python3
                pass

        if not shebang or not shebang.startswith('#!'):
            raise OSError(
                'Unable to make {} as executable by chmod '
                'and detect interpreter from shebang.'.format(filepath))
        ret = shebang[2:].strip().split() + [filepath]
    return ret


def brief_list(blist, base=0):
    """@API
    Briefly show an integer list, combine the continuous numbers.
    @params:
        blist: The list
    @returns:
        (str): The string to show for the briefed list.
    """
    if not blist:
        return "[]"
    blist = [b + base for b in blist]
    if len(blist) == 1:
        return str(blist[0])
    blist = sorted(blist)
    groups = [[]]
    ret = []
    for i in range(0, len(blist) - 1):
        ele0 = blist[i]
        ele1 = blist[i + 1]
        if ele1 - ele0 > 1:
            groups[-1].append(ele0)
            groups.append([])
        else:
            groups[-1].append(ele0)
    groups[-1].append(blist[-1])
    for group in groups:
        if len(group) == 1:
            ret.append(str(group[0]))
        elif len(group) == 2:
            ret.append(str(group[0]))
            ret.append(str(group[1]))
        else:
            ret.append(str(group[0]) + '-' + str(group[-1]))
    return ', '.join(ret)


class ThreadEx(Thread):
    """
    A thread able to send exception to main thread
    thread.ex will hold the exception.
    """
    def __init__(self,
                 group=None,
                 target=None,
                 name=None,
                 args=None,
                 kwargs=None):
        # pylint: disable=too-many-arguments
        Thread.__init__(self, group, target, name, args or (), kwargs or {})
        self.daemon = True
        self.ex = None

    def run(self):
        try:
            Thread.run(self)
        except Exception as ex:  # pylint: disable=broad-except
            #from traceback import format_exc
            self.ex = ex


class ThreadPool:  # pylint: disable=too-few-public-methods
    """
    A thread manager for ThreadEx.
    """
    def __init__(self, nthread, initializer=None, initargs=None):
        self.threads = []
        if not isinstance(initargs, list):
            initargs = [(initargs, ) if initargs else ()] * nthread
        for i in range(nthread):
            thread = ThreadEx(target=initializer, args=initargs[i])
            thread.start()
            self.threads.append(thread)

    def join(self, interval=1, cleanup=None):
        """
        Try to join the threads, able to respond to KeyboardInterrupt
        @params:
            `interval`: The interval/timeout to join every time.
            `cleanup` : The cleanup function
        """
        try:
            while True:
                threads_alive = 0
                for thread in self.threads:
                    # check if the thread is done
                    thread.join(timeout=interval)
                    if thread.ex:
                        # exception raised, try to quit and cleanup
                        if not callable(cleanup):
                            raise thread.ex
                        cleanup(thread.ex)
                        threads_alive = 0
                        break
                    if thread.is_alive():
                        threads_alive += 1
                if threads_alive == 0:
                    break
        except KeyboardInterrupt as ex:  # pragma: no cover
            if callable(cleanup):
                cleanup(ex=ex)


class PQueue(PriorityQueue):
    """@API
    A modified PriorityQueue, which allows jobs to be submitted in batch
    """

    # pylint: disable=arguments-differ
    def __init__(self, maxsize=0, batch_len=None):
        """@API
        A Priority Queue for PyPPL jobs
            0                             0 done,           wait for 1
              1       start 0    1        start 1           start 2
                2     ------>  0   2      ------>    2      --------->
                  3                  3             1   3                    3
                    4                  4                 4                2   4
                                                                        1
        @params:
            maxsize  : The maxsize of the queue. Default: None
            batch_len: What's the length of a batch
        """
        if not batch_len:
            raise ValueError('`batch_len` is required for PQueue.')
        PriorityQueue.__init__(self, maxsize)
        self.batch_len = batch_len

    def put_next(self, item, batch):
        """@API
        Put item to next batch
        @params:
            item (any): item to put
            batch (int): current batch
        """
        self.put(item, batch + 2)

    def put(self, item, batch=None):
        """@API
        Put item to any batch
        @params:
            item (any): item to put
            batch (int): target batch
        """
        batch = batch or item
        PriorityQueue.put(self, item + batch * self.batch_len)

    def get(self):
        """@API
        Get an item from the queue
        @returns:
            (int, int): The index of the item and the batch of it
        """
        item = PriorityQueue.get(self)
        batch, index = divmod(item, self.batch_len)
        return index, batch


class _MultiDestTransition(Transition):
    """Transition with multiple destination"""

    # pylint: disable=too-many-arguments
    def __init__(self,
                 source,
                 dest,
                 conditions=None,
                 unless=None,
                 before=None,
                 after=None,
                 prepare=None,
                 **kwargs):
        self._result = self._dest = None
        super().__init__(source, dest, conditions, unless, before, after,
                         prepare)
        if isinstance(dest, dict):
            self._func = kwargs.pop('depends_on', None)
            if not self._func:
                raise AttributeError(
                    "A multi-destination transition requires a 'depends_on'")
        else:
            # use base version in case transition does not need special handling
            self.execute = super().execute

    def execute(self, event_data):  # pylint: disable=method-hidden
        """Excute the function"""
        func = self._func if callable(self._func) else getattr(
            event_data.model, self._func)
        self._result = func()
        super().execute(event_data)

    @property
    def dest(self):
        """Get the destination"""
        return self._dest[
            self._result] if self._result is not None else self._dest

    @dest.setter
    def dest(self, value):
        self._dest = value


class StateMachine(Machine):  # pylint: disable=too-few-public-methods
    """StateMachine with multiple destination support"""
    transition_cls = _MultiDestTransition



"""
Make runners as plugins for PyPPL
@variables:
    RMNAME (str): The name of the runner manager
    RUNNERS (dict): All ever registered runners
    DEFAULT_POLL_INTERVAL (int): The default poll interval to check job status
    hookimpl (pluggy.HookimplMarker): The marker for runner hook Implementations
    hookspec (pluggy.HookspecMarker): The marker for runner hooks
"""
# pylint: disable=unused-argument,invalid-name
import sys
import pluggy
import psutil
from .exception import (RunnerNoSuchRunner, RunnerMorethanOneRunnerEnabled,
                        RunnerTypeError)

RMNAME = "pyppl_runner"
# save all runners ever registered
RUNNERS = {}
# poll interval to check job status
DEFAULT_POLL_INTERVAL = 1

hookimpl = pluggy.HookimplMarker(RMNAME)
hookspec = pluggy.HookspecMarker(RMNAME)


@hookspec
def runner_init(proc):
    """@API
    RUNNER API
    Initiate runner
    @params:
        proc (Proc): The Proc instance
    """


@hookspec(firstresult=True)
def isrunning(job):
    """@API
    RUNNER API
    Tell if the job is running
    @params:
        job (Job): the job instance
    """


@hookspec(firstresult=True)
def submit(job):
    """@API
    RUNNER API
    Submit a job
    @params:
        job (Job): the job instance
    """


@hookspec(firstresult=True)
def kill(job):
    """@API
    RUNNER API
    Try to kill the job
    @params:
        job (Job): the job instance
    """


@hookspec(firstresult=True)
def script_parts(job, base):
    """@API
    RUNNER API
    Overwrite script parts
    @params:
        job (Job): the job instance
        base (Diot): The base script parts
    """


runnermgr = pluggy.PluginManager(RMNAME)
runnermgr.add_hookspecs(sys.modules[__name__])


def _runner_name(runner):
    if hasattr(runner, '__name__'):
        name = runner.__name__
    else:
        name = runner.__class__.__name__

    if name[:13] in ('pyppl_runner_', 'pyppl-runner-'):
        name = name[13:]
    elif name[:11] == 'PyPPLRunner':
        name = name[11:].lower()
    return name


def use_runner(runner):
    """@API
    runner should be a module or the name of a module,
    with or without "pyppl_runner_" prefix
    To enable a runner, we need to disable other runners
    @params:
        runner (str): the name of runner
    """
    if runner not in RUNNERS:
        raise RunnerNoSuchRunner(runner)
    for plugin in RUNNERS.values():
        if _runner_name(plugin) == runner:
            if not runnermgr.is_registered(plugin):
                runnermgr.register(plugin)
        elif runnermgr.is_registered(plugin):
            runnermgr.unregister(plugin)
    assert len(runnermgr.get_plugins()) == 1, (
        'One runner is allow at a time. We have {}'.format(
            runnermgr.get_plugins()
        )
    )


def current_runner():
    """@API
    Get current runner name
    @returns:
        (str): current runner name
    """
    all_runners = list(runnermgr.get_plugins())
    if len(all_runners) > 1:
        raise RunnerMorethanOneRunnerEnabled(str(all_runners))
    if not all_runners:
        raise RunnerNoSuchRunner('No runners registered.')

    for name, plugin in RUNNERS.items():
        if plugin is all_runners[0]:
            return name
    raise RunnerNoSuchRunner('No runners registered.')


def poll_interval():
    """@API
    Get the poll interval for current runner
    @returns:
        (int): poll interval for querying job status
    """
    runner = RUNNERS[current_runner()]
    if hasattr(runner, 'POLL_INTERVAL'):
        return int(runner.POLL_INTERVAL)
    return DEFAULT_POLL_INTERVAL


def register_runner(runner, name=None):
    """@API
    Register a runner
    @params:
        runner (callable): The runner, a module or a class object
        name (str): The name of the runner to registered
    """
    name = name or _runner_name(runner)
    if not runnermgr.is_registered(runner):
        # we only allow an object from one class to be registered once
        for plugin in runnermgr.get_plugins():
            if _runner_name(plugin) == name:
                raise RunnerTypeError(
                    'Runner %r has already been registered.' % (name))
        runnermgr.register(runner, name=name)
    RUNNERS[name] = runner

def killtree(pid, killme=True, signal=9):
    """Kill a process tree"""
    myself = psutil.Process(int(pid))
    children = myself.children(recursive=True)
    if killme:
        children.append(myself)
    for proc in children:
        proc.send_signal(signal)

    return bool(psutil.wait_procs(children))

class PyPPLRunnerLocal:
    """@API
    PyPPL's default runner"""
    # pylint: disable=no-self-use

    __version__ = 'builtin'

    @hookimpl
    def kill(self, job):
        """@API
        Try to kill the running jobs if I am exiting
        @params:
            job (Job): the job instance
        """
        if int(job.pid) > 0:
            return killtree(job.pid)
        return True

    @hookimpl
    def submit(self, job):
        """@API
        Try to submit the job
        @params:
            job (Job): the job instance
        """
        import cmdy
        cmd = cmdy._(*job.script[1:], _exe=job.script[0], _raise=False).iter
        cmd._rc = 0
        job.pid = cmd.pid
        return cmd

    @hookimpl
    def isrunning(self, job):
        """@API
        Try to tell whether the job is still running.
        @params:
            job (Job): the job instance
        @returns:
            `True` if yes, otherwise `False`
        """
        if not job.pid or int(job.pid) < 0:
            return False
        return psutil.pid_exists(int(job.pid))


register_runner(PyPPLRunnerLocal())



"""PyPPL - A Python PiPeLine framework."""

__version__ = "3.2.2"

from .config import load_config
from .plugin import config_plugins
from .pyppl import PyPPL, PROCESSES, PIPELINES
from .proc import Proc
from .procset import ProcSet
from .channel import Channel
from .runner import register_runner



"""Pipeline for PyPPL
@variables:
    SEPARATOR_LEN (int): the length of the separators in log
    PROCESSES (set): The process pool where the processes are registered
    TIPS (list): Some tips to show in log
    PIPELINES (dict): Exists pipelines
"""
import sys
# give random tips in the log
import random
import textwrap
import fnmatch
from pathlib import Path
from diot import Diot
from simpleconf import Config
from .config import config as default_config, DEFAULT_CFGFILES, _config_factory
from .logger import logger
from .plugin import pluginmgr, config_plugins
from .runner import RUNNERS
from .exception import (PyPPLInvalidConfigurationKey,
                        PyPPLNameError,
                        ProcessAlreadyRegistered,
                        PyPPLWrongPositionMethod,
                        PyPPLMethodExists)
from .utils import try_deepcopy, name2filename, fs, log_msg_len

# length of separators in log
SEPARATOR_LEN = log_msg_len(ret='with-nothing', minlen=45)
# Regiester processes
PROCESSES = set()
# tips
TIPS = [
    "You can find the stdout in <workdir>/<job.index>/job.stdout",
    "You can find the stderr in <workdir>/<job.index>/job.stderr",
    "You can find the script in <workdir>/<job.index>/job.script",
    "Check documentation at: https://pyppl.readthedocs.io/en/latest/",
    "You cannot have two processes with the same id and tag",
    "Workdir defaults to PyPPL.<id>.<tag>.<suffix> under default <ppldir>",
    "The default <ppldir> is './workdir'"
]

# name of defined pipelines
PIPELINES = {}

def _register_proc(proc):
    """Register process into the pool"""
    for registered_proc in PROCESSES:
        if registered_proc.name == proc.name and registered_proc is not proc:
            raise ProcessAlreadyRegistered(proc1=registered_proc,
                                           proc2=proc) from None
    PROCESSES.add(proc)


def _get_next_procs(procs):
    """Get next possible processes to run"""
    nextprocs = [
        nextproc
        for proc in procs
        if proc.nexts
        for nextproc in proc.nexts
        if (nextproc not in procs
            and all(depend in procs
                    for depend in nextproc.depends))
    ]
    ret = []
    for nproc in nextprocs:
        if nproc not in ret:
            ret.append(nproc)
    return ret


def _anything2procs(*anything, procset='starts'):
    """Translate anything to a list of processes
    It actually serves as a process selector.
    Keep in mind that the order of the processes is not guaranteed.
    """
    from .proc import Proc
    from .procset import ProcSet, Proxy
    ret = Proxy()
    for anyth in anything:
        if isinstance(anyth, Proc):
            ret.add(anyth)
        elif isinstance(anyth, ProcSet):
            ret.add(anyth.starts if procset == 'starts' else anyth.ends)
        elif isinstance(anyth, (tuple, list)):
            for thing in anyth:
                ret.add(_anything2procs(thing, procset=procset))
        else:
            for proc in PROCESSES:
                if anyth in (proc.id, proc.shortname, proc.name):
                    ret.add(proc)
                elif fnmatch.fnmatch(proc.shortname, anyth):
                    ret.add(proc)
    return ret


def _logo():
    from . import __version__
    logger.pyppl('+' + r'-' * (SEPARATOR_LEN - 2) + '+')
    for logo_line in (r'',
                      r'________        ______________________ ',
                      r'___  __ \____  ____  __ \__  __ \__  / ',
                      r'__  /_/ /_  / / /_  /_/ /_  /_/ /_  /  ',
                      r'_  ____/_  /_/ /_  ____/_  ____/_  /___',
                      r'/_/     _\__, / /_/     /_/     /_____/',
                      r'        /____/                         ',
                      r'',
                      r'v{}'.format(__version__),
                      r''):
        logger.pyppl('|%s|' % logo_line.center(SEPARATOR_LEN - 2))

    logger.pyppl('+' + r'-' * (SEPARATOR_LEN - 2) + '+')

    logger.tips(random.choice(TIPS))

    for cfgfile in DEFAULT_CFGFILES:
        if fs.exists(cfgfile) or str(cfgfile).endswith('.osenv'):
            logger.config('Read from %s', cfgfile)
    plugins_to_log = ((name[6:] if name[:6] in ('pyppl_', 'pyppl-',
                                                'pyppl.') else name,
                       plugin.__version__
                       if hasattr(plugin, '__version__') else 'unknown')
                      for name, plugin in pluginmgr.list_name_plugin())
    plugins_to_log = sorted(plugins_to_log,
                            key=lambda nver: nver[1] != 'builtin')
    plugins_to_log = ', '.join('-'.join(namever) for namever in plugins_to_log)
    plugins_to_log = textwrap.wrap(plugins_to_log,
                                   initial_indent='  ',
                                   subsequent_indent='  ',
                                   break_long_words=False,
                                   break_on_hyphens=False)
    logger.plugin('Loaded plugins:')
    for plugin_to_log in plugins_to_log:
        logger.plugin(plugin_to_log)

    runners_to_log = ((rname, rplugin.__version__ if hasattr(
        rplugin, '__version__') else 'unknown')
                      for rname, rplugin in RUNNERS.items())
    runners_to_log = sorted(runners_to_log,
                            key=lambda nver: nver[1] != 'builtin')
    runners_to_log = ', '.join('-'.join(namever) for namever in runners_to_log)
    runners_to_log = textwrap.wrap(runners_to_log,
                                   initial_indent='  ',
                                   subsequent_indent='  ',
                                   break_long_words=False,
                                   break_on_hyphens=False)
    logger.plugin('Loaded runners:')
    for runner_to_log in runners_to_log:
        logger.plugin(runner_to_log)


def _parse_kwconfigs(kwconfigs):
    """Allow logger = {'level': 'debug'} to be
    specified as logger_level = 'debug'"""
    ret = {}
    if isinstance(kwconfigs.get('runner'), str):
        ret['runner'] = dict(runner=kwconfigs.pop('runner'))
    for key, val in kwconfigs.items():
        if key[:7] == 'logger_':
            ret.setdefault('logger', {})[key[7:]] = val
        elif key[:7] == 'runner_':
            ret.setdefault('runner', {})[key[7:]] = val
        elif key[:5] == 'envs_':
            ret.setdefault('envs', {})[key[5:]] = val
        elif key[:7] == 'config_':
            ret.setdefault('config', {})[key[7:]] = val
        else:
            ret[key] = val
    return ret

class PyPPL:
    """@API
    The class for the whole pipeline
    """
    __slots__ = ('name',
                 'runtime_config',
                 'procs',
                 'props',
                 'starts',
                 'ends',
                 '__dict__') # allow adding methods

    def __init__(self, config=None, name=None, config_files=None, **kwconfigs):
        """@API
        The construct for PyPPL
        @params:
            config (dict): the runtime configuration for the pipeline
            name (str): The name of the pipeline
            config_files (list): A list of runtime configuration files
            **kwconfigs: flattened runtime configurations, for example
                - you can do: `PyPPL(forks = 10)`, or even
                - `PyPPL(logger_level = 'debug')`
        """
        kwconfigs = _parse_kwconfigs(kwconfigs)
        # check if keys in kwconfigs are valid
        config = config or {}
        config.update(kwconfigs)
        for key in config:
            if key not in default_config:
                raise PyPPLInvalidConfigurationKey(
                    'No such configuration key: ' + key)

        self.name = (name2filename(name)
                     if name
                     else Path(sys.argv[0]).stem
                     if not PIPELINES
                     else Path(sys.argv[0]).stem + str(len(PIPELINES) + 1))
        if self.name in PIPELINES:
            raise PyPPLNameError(
                'Pipeline name {!r}({!r}) has been used.'.format(
                    self.name, name))
        PIPELINES[self.name] = self

        if config_files and not isinstance(config_files, (tuple, list)):
            config_files = [config_files]

        self.runtime_config = Config()
        self.runtime_config._load(*(config_files or ()))
        self.runtime_config._load(dict(default=config))

        logger_config = try_deepcopy(default_config.logger.dict())
        logger_config.update(self.runtime_config.pop('logger', {}))
        if logger_config['file'] is True:
            logger_config['file'] = './{}.pyppl.log'.format(self.name)
        logger.initialize(logger_config)
        del logger_config

        config = self.runtime_config.pop('plugins', [])
        config_plugins(*config)

        _logo()
        logger.pyppl('=' * SEPARATOR_LEN)
        logger.pyppl('>>> PIPELINE: {}'.format(self.name))
        logger.pyppl('=' * SEPARATOR_LEN)

        self.procs = []
        self.starts = []
        self.ends = []
        self.props = Diot()  # for plugins
        pluginmgr.hook.pyppl_init(ppl=self)

    def run(self, profile='default'): # pylint: disable=too-many-branches
        """@API
        Run the pipeline with certain profile
        @params:
            profile (str): The profile name
        """
        # This function needs to be simplified later

        # for default profile, we shall not load anything from default_config
        # as part/base of runtime config, since they have alread been used as
        # default values when initialize Proc object
        defconfig = Config()
        if profile == 'default':
            pass
        elif profile in default_config._profiles:
            # Otherwise, we just use the settings of profile from default_config
            # as the base of runtime config
            with default_config._with(profile, copy=True) as dconfig:
                defconfig._load({profile: dconfig.dict()})
        else: # assuming direct runner
            defconfig._load({'default': {'runner': profile}},
                            factory=_config_factory)

        defconfig._load(self.runtime_config)
        # make available the direct runner
        if profile not in defconfig._profiles:
            defconfig._load({profile: dict(runner=profile)},
                            factory=_config_factory)

        # use the profile, based on default profile
        defconfig._use(profile, 'default')
        # will not be used for procs
        defconfig.pop('logger', None)
        defconfig.pop('plugins', None)

        # update tags, and maybe someother props that are import in prerun
        if 'tag' in defconfig:
            for proc in self.procs:
                if proc._setcounter.get('tag', 0) > 0:
                    continue
                proc.tag = defconfig.tag

        # plugins can stop pipeling being running
        if pluginmgr.hook.pyppl_prerun(ppl=self) is False:
            pluginmgr.hook.pyppl_postrun(ppl=self)
            return self

        try:
            for proc in self.procs:
                # echo the process name and description
                name = '%s%s: ' % (
                    proc.name,
                    ' (%s)' % proc.origin
                    if proc.origin and proc.origin != proc.id
                    else ''
                )

                logger.process('-' * SEPARATOR_LEN)
                if len(name + proc.desc) > SEPARATOR_LEN:
                    logger.process(name)
                    for i in textwrap.wrap(proc.desc,
                                           SEPARATOR_LEN,
                                           initial_indent='  ',
                                           subsequent_indent='  '):
                        logger.process(i)
                else:
                    logger.process(name + proc.desc)

                logger.process('-' * SEPARATOR_LEN)

                # echo the dependencies
                depends = ([dproc.name for dproc in proc.depends]
                           if proc.depends
                           else ['START'])
                nexts = ([nproc.name for nproc in proc.nexts]
                         if proc.nexts
                         else ['END'])
                depmaxlen = max([len(dep) for dep in depends])
                nxtmaxlen = max([len(nxt) for nxt in nexts])
                lendiff = len(depends) - len(nexts)
                lessprocs = depends if lendiff < 0 else nexts
                lendiff = abs(lendiff)
                lessprocs.extend([''] * (lendiff - int(lendiff / 2)))
                for i in range(int(lendiff/2)):
                    lessprocs.insert(0, '')

                for i in range(len(lessprocs)):
                    middle = (proc.name if i == int((len(lessprocs) - 1) / 2)
                              else ' ' * len(proc.name))
                    logger.depends(
                        '| {0} | {1} {2} {1} | {3} |'.format(
                            depends[i].ljust(depmaxlen),
                            '=>' if i == int((len(lessprocs)-1)/2) else '  ',
                            middle,
                            nexts[i].ljust(nxtmaxlen)
                        )
                    )

                proc.run(defconfig)

        except BaseException as ex:
            logger.error(f"{type(ex).__name__}:")
            logger.error(str(ex))
            if pluginmgr.hook.pyppl_postrun(ppl=self) is not False:
                raise
        else:
            pluginmgr.hook.pyppl_postrun(ppl=self)

        return self

    def start(self, *anything):
        """@API
        Set the start processes for the pipeline
        @params:
            *anything: Anything that can be converted to processes
                - Could be a string or a wildcard to search for processes
                - or the process itself
        """
        del self.procs[:]
        del self.starts[:]
        del self.ends[:]

        starts = _anything2procs(*anything)
        # Let's check if start process depending on others
        for start in starts:
            if start.depends:
                logger.warning(
                    "Start process %r ignored, as it's depending on:",
                    start.name
                )
                logger.warning('  %s' % start.depends)
            else:
                self.starts.append(start)

        self.procs = self.starts[:]
        # fetch all procs depending on starts in the order
        # that they will be excuted.
        nexts = _get_next_procs(self.procs)
        if not nexts:
            for proc in self.procs:
                self.ends.append(proc)
                if proc.nexts:
                    logger.warning(
                        "%r will not run, as they depend on "
                        "non-start processes or themselves.",
                        proc.nexts
                    )
            return self

        while nexts:
            for nextproc in nexts:
                self.procs.append(nextproc)
                if not nextproc.nexts:
                    self.ends.append(nextproc)
            nexts = _get_next_procs(self.procs)
        return self

    def add_method(self, func, require='start'):
        """@API
        Add a method to PyPPL object, used as a decorator
        @params:
            func (callable): the function to add
            require (str): Require which function to be called before this
        """
        name = func.__name__

        if hasattr(self, name):
            raise PyPPLMethodExists(
                'Method name %r has already been registered.' % name)

        def wrapper(*args, **kwargs):
            if require == 'start' and not self.starts:
                raise PyPPLWrongPositionMethod(
                    '%s requires .start() to be called.' % name)

            if require == 'run' and (not self.procs
                                     or not self.procs[0].channel):
                raise PyPPLWrongPositionMethod(
                    '%s requires .run() to be called.' % name)

            func(self, *args, **kwargs)
            return self

        self.__dict__[name] = wrapper



"""Show PyPPL logo, version and loaded plugins and runners,
based on default configurations."""
from ..plugin import hookimpl
from ..pyppl import _logo


@hookimpl
def cli_addcmd(commands):
    """Add logo command"""
    commands.logo._hbald = False
    commands.logo = __doc__


@hookimpl
def cli_execcmd(command, opts):  # pylint: disable=unused-argument
    """Run the command"""
    if command == 'logo':
        _logo()



"""Configuration for PyPPL
@variables:
    DEFAULT_CFGFILES (tuple): default configuration files
    DEFAULT_CONFIG (dict): default configurations to be loaded
    config (dict): The default configuration that will be
        used in the whole session
"""
from multiprocessing import cpu_count
from pathlib import Path
from diot import Diot
from simpleconf import Config
from .plugin import config_plugins, pluginmgr, PMNAME
from .runner import runnermgr, RMNAME, register_runner

# priority: lowest -> highest
DEFAULT_CFGFILES = (Path('~/.PyPPL.toml').expanduser(), Path('./.PyPPL.toml'),
                    'PYPPL.osenv')

# default configurations
DEFAULT_CONFIG = dict(default=dict(
    # plugins
    plugins=[],
    # logger options
    logger=dict(file=None, theme='green_on_black', level='info',
                leveldiffs=[]),
    # The cache option, True/False/export
    cache=True,
    # Whether expand directory to check signature
    dirsig=True,
    # How to deal with the errors
    # retry, ignore, halt
    # halt to halt the whole pipeline, no submitting new jobs
    # terminate to just terminate the job itself
    errhow='terminate',
    # How many times to retry to jobs once error occurs
    errntry=3,
    # The directory to export the output files
    forks=1,
    # Default shell/language
    lang='bash',
    # number of threads used to build jobs and to check job cache status
    nthread=min(int(cpu_count() / 2), 16),
    # Where cache file and workdir located
    ppldir='./workdir',
    # Select the runner
    runner='local',
    # The tag of the job
    tag='notag',
    # The template engine (name)
    template='liquid',
    # The template environment
    envs=Diot(),
    # configs for plugins
    config={},
))

config = Config()  # pylint: disable=invalid-name

def _config_factory(conf):
    """The factory to convert runner to a dict"""
    # we have profiles
    for _, con in conf.items():
        if 'runner' not in con or isinstance(con['runner'], dict):
            continue
        # you can't set profile in config, so it's direct runner
        con['runner'] = {'runner': con['runner']}
    return conf

def load_config(default_config, *config_files):
    """@API
    Load the configurations
    @params:
        default_config (dict|path): A configuration dictionary or
              a path to a configuration file
        *config_files: A list of configuration or configuration files
    """
    config.clear()
    config._load(DEFAULT_CONFIG, default_config, *config_files,
                 factory=_config_factory)

load_config(*DEFAULT_CFGFILES)
pluginmgr.load_setuptools_entrypoints(PMNAME)
runnermgr.load_setuptools_entrypoints(RMNAME)
for runner in runnermgr.get_plugins():
    # save runners in memory
    register_runner(runner)
config_plugins(*config.plugins)
# deprecate this, since at this point, no plugins are loaded yet
pluginmgr.hook.setup(config=config)



"""List available profiles."""
import toml
from ..plugin import hookimpl
from ..config import config
from ..logger import logger as plogger

@hookimpl
def logger_init(logger):
    """Add log levels"""
    config.logger.level = 'debug'
    logger.add_level('PROFILE', 'TITLE')
    logger.add_level('-', 'DEBUG')
    logger.add_level('=', 'WARNING')


@hookimpl
def cli_addcmd(commands):
    """Add profile command"""
    commands.profile._hbald = False
    commands.profile = __doc__
    commands.profile.name = ''
    commands.profile.name.desc = 'The name of the profile to show'
    commands.profile.nodefault = False
    commands.profile.nodefault.desc = 'Hide default configuration items or not.'
    commands.runner = commands.profile

def _get_defaults():
    config._use('default')
    defaults = config.dict()
    for _, cached in config._protected['cached'].items():
        if 'default' in cached:
            del cached['default']

    return defaults

def _show_profile(profile, defaults, nodefault):
    plogger.profile('>>> %s', profile)
    config._use(profile)
    confs = config.dict() if profile != 'default' else defaults
    conflines = toml.dumps(confs).splitlines()
    if nodefault:
        for line in conflines:
            plogger['=']('    %s', line)
    else:
        defaults = defaults.copy()
        defaults.update(confs)
        for line in toml.dumps(defaults).splitlines():
            if line in conflines or profile == 'default':
                plogger['=']('    %s' % line)
            else:
                plogger['-']('    %s' % line)

@hookimpl
def cli_execcmd(command, opts):  # pylint: disable=unused-argument
    """Run the command"""
    if command in ('runner', 'profile'):
        defaults = _get_defaults()

        if opts.name:
            _show_profile(opts.name, defaults, opts.nodefault)
            return

        for profile in sorted(config._profiles,
                              key=lambda prof: ('_'
                                                if prof == 'default'
                                                else prof)):
            _show_profile(profile, defaults, opts.nodefault)



"""Command line tool for PyPPL"""
import sys
import importlib
from pathlib import Path
from diot import Diot
from pyparam import commands
from ..logger import logger
from ..plugin import config_plugins, pluginmgr
from .. import __version__

CONSOLE_PLUGINS = [
    importlib.import_module('.' + cmdfile.stem, __name__)
    for cmdfile in Path(__file__).parent.glob('*.py')
    if cmdfile.stem[:1] != '_'
]
for console_plugin in CONSOLE_PLUGINS:
    console_plugin.__version__ = 'builtin'

config_plugins(*CONSOLE_PLUGINS)

commands._desc = 'Command Line Tool for PyPPL v{}'.format(__version__)
pluginmgr.hook.cli_addcmd(commands=commands)


def main():
    """Entry point"""
    command, opts, _ = commands._parse(dict_wrapper=Diot)
    pluginmgr.hook.logger_init(logger=logger)
    logger.initialize()
    pluginmgr.hook.cli_execcmd(command=command, opts=opts)



"""Implementations for Proc"""

import sys
import uuid
import textwrap
from pathlib import Path
from liquid.stream import safe_split
from diot import OrderedDiot, Diot
import cmdy
from simpleconf import Config
from . import template
from .runner import use_runner
from .config import config
from .exception import (ProcessAttributeError,
                        ProcessInputError,
                        ProcessOutputError,
                        ProcessScriptError)
from .channel import Channel
from .logger import logger
from .utils import funcsig, always_list, fs
from .plugin import PluginConfig

# pylint: disable=unused-argument

OUT_VARTYPE = ['var']
OUT_FILETYPE = ['file', 'path']
OUT_DIRTYPE = ['dir', 'folder']
OUT_STDOUTTYPE = ['stdout']
OUT_STDERRTYPE = ['stderr']

IN_VARTYPE = ['var']
IN_FILETYPE = ['file', 'path', 'dir', 'folder']
IN_FILESTYPE = ['files', 'paths', 'dirs', 'folders']

def _require(this, name, strict=True, msg=None):
    """
    strict:
        True: any falsy values not allowed
        False: only None not allowed
    """
    if not hasattr(this, name):
        raise ProcessAttributeError(
            msg or
            'Process has no such attribute {!r}.'.format(name)
        )
    value = getattr(this, name)
    if (strict and not value) or value is None:
        raise ProcessAttributeError(
            msg or
            'Process attribute {!r} has not been initialized.'.format(name)
        )

def _decache(this, name):
    if name in this.__attrs_property_cached__:
        del this.__attrs_property_cached__[name]

def proc_setter_count(this, value=None, name=None):
    """Counter for some properties
    in order to prevent overwriting by runtime_config if
    they are set by `pXXX.dirsig = True`
    """
    this._setcounter.setdefault(name, -1)
    # set it to 0 in __init__
    this._setcounter[name] += 1

def proc_runtime_config_setter(this, value):
    """Try to override the proc property values
    if possible from runtime config"""
    # replace the values in runtime config
    if not value:
        return
    for key, val in value.items():
        if key == 'envs':
            this.envs.update(val)
        elif key == 'config':
            this.config.update(val)
        elif this._setcounter.get(key) or key == 'runner':
            continue
        else:
            setattr(this, key, val)

def proc_config_setter(this, value):
    """Try to update config whever a config is passed"""
    preval = this.__attrs_property_raw__.get('config')
    if not preval:
        default = PluginConfig(config.config)
        default.update(value or {})
        return default
    preval.update(value or {})
    return preval

def proc_id_setter(this, value):
    """Don't allow id to be set using setter"""
    if this._setcounter.get('id'):
        raise ProcessAttributeError(
            'Process id should be passed while instantization.'
        )
    proc_tag_setter(this, value, name='id')
    this._setcounter['id'] += 1

def proc_tag_setter(this, value, name='tag'):
    """Tag setter"""
    this._setcounter.setdefault(name, -1)
    this._setcounter[name] += 1
    _decache(this, 'name')
    _decache(this, 'shortname')
    _decache(this, 'procset')

def proc_procset(this, value):
    """Parse the procset from tag"""
    _require(this,
             'tag',
             strict=False,
             msg='Process procset requires tag to be specified.')
    if '@' not in this.tag:
        return ''
    return this.tag.split('@', 1)[1]

def proc_name(this, value):
    """Get the name of the process, including all parts of tag"""
    return '{}.{}'.format(this.id, this.tag)

def proc_shortname(this, value):
    """Get the shortname of the process,
    if tag starts with "notag", don't include it"""
    if not this.tag or this.tag == 'notag':
        return this.id
    if this.tag[:6] == 'notag@':
        return '{}@{}'.format(this.id, this.tag[6:])
    return this.name

def proc_runner(this, value):
    """Turn a runner profile to a runner config
    If it is a config, update the default one with it,
    otherwise, read the runner config from profile
    if value is a dict given by proc.runner = {...}, then we should make it
    based on defaults (both config and runtime_config), otherwise, we should
    check if it is a profile or a direct runner name
    """
    runtime_config = this.runtime_config or Config()
    # Use runtime_config's runner if not set. default_config.runner
    # should be ignored when initialized
    if (not this._setcounter.get('runner')
            and this.runtime_config.get('runner')):
        value = this.runtime_config.runner

    default_runner = {}
    with config._with('default', copy=True) as dconf:
        default_runner.update(dconf.get('runner', {}))
    with runtime_config._with('default', copy=True) as dconf:
        default_runner.update(dconf.get('runner', {}))

    if isinstance(value, dict):
        default_runner.update(value)
    elif value in config._profiles:
        with config._with(value, copy=True) as vconf:
            default_runner.update(vconf.get('runner', {}))
    elif value in runtime_config._profiles:
        with runtime_config._with(value, copy=True) as vconf:
            default_runner.update(vconf.get('runner', {}))
    else:
        default_runner['runner'] = value

    default_runner.setdefault('runner', 'local')
    # in case we have 'sge.runner', we turn it into 'sge_runner'
    # have to copy the keys, otherwise, dictionary changed size during
    # iteration
    for key in list(default_runner):
        default_runner[key.replace('.', '_')] = default_runner.pop(key)
    return Diot(default_runner)

def proc_depends_setter(this, value):
    """Try to convert all possible dependencies to processes"""
    from .pyppl import _anything2procs
    depends = _anything2procs(value, procset='ends')
    try:
        prev_depends = this._depends
    except KeyError:
        prev_depends = []
    # remove process in previous depends' nexts
    for prevdep in prev_depends:
        prevdep.nexts.remove(this)
    for depend in depends:
        depend.nexts.append(this)
    return depends

def proc_input_setter(this, value):
    """Allow input to be set twice:
    proc.input = 'a'
    proc.input = [1]
    Then real input would be:
    proc.input = {'a': [1]}
    """
    # reset the size and jobs
    _decache(this, 'size')
    _decache(this, 'jobs')
    _decache(this, 'suffix')
    if isinstance(value, (dict, str)):
        # ignore previous input
        return value
    try:
        prev_input = this._input
    except KeyError:
        return value
    if isinstance(prev_input, str):
        return {prev_input: value}
    # prev_input not a str, ignore it
    if isinstance(value, (list, tuple)):
        return ', '.join(value)
    raise ProcessInputError(
        'Unsupported input format, '
        'expecting str, list, OrderedDict, diot.OrderedDiot '
        'or dict with a sole key-value pair.'
    )

def proc_input(this, value): # pylint: disable=too-many-locals,too-many-branches
    """Parse and prepare the input, allowing jobs to easily access it"""
    # parse this.config.input keys
    # even for skipped or resumed process
    # because we need to keep the order
    # however, yaml does not
    # ['a:var', 'b:file', ...]

    # make sure it's called after .run()
    _require(this,
             'runtime_config',
             strict=False,
             msg='Process input should be accessed after it starts running')
    input_keys_and_types = (sum((always_list(key) for key in value), [])
                            if isinstance(value, dict)
                            else always_list(value)
                            if value
                            else [])

    # {'a': 'var', 'b': 'file', ...}
    input_keytypes = OrderedDiot()
    for keytype in input_keys_and_types:
        if not keytype.strip():
            continue
        if ':' not in keytype:
            input_keytypes[keytype] = IN_VARTYPE[0]
        else:
            thekey, thetype = keytype.split(':', 1)
            if thetype not in IN_VARTYPE + IN_FILESTYPE + IN_FILETYPE:
                del this.__attrs_property_raw__['input']
                raise ProcessInputError('Unknown input type: ' + thetype)
            input_keytypes[thekey] = thetype
    del input_keys_and_types

    ret = OrderedDiot()
    # no data specified, inherit from depends or argv
    input_values = (list(value.values())
                    if isinstance(value, dict)
                    else [Channel.from_channels(*[d.channel
                                                  for d in this.depends])
                          if this.depends
                          else Channel.from_argv()])

    input_channel = Channel.create()
    for invalue in input_values:
        # a callback, on all channels
        if callable(invalue):
            input_channel = input_channel.cbind(
                invalue(*[d.channel for d in this.depends]
                        if this.depends else Channel.from_argv())
            )
        elif isinstance(invalue, Channel):
            input_channel = input_channel.cbind(invalue)
        else:
            input_channel = input_channel.cbind(Channel.create(invalue))
    data_width = input_channel.width()
    key_width = len(input_keytypes)
    if key_width < data_width:
        logger.warning('Not all data are used as input, %s column(s) wasted.',
                       (data_width - key_width),
                       proc=this.id)
    # compose ret
    for i, inkey in enumerate(input_keytypes):
        intype = input_keytypes[inkey]
        ret[inkey] = (intype, [])

        if i >= data_width:
            if intype in IN_FILESTYPE:
                ret[inkey][1].extend([[]] * input_channel.length())
            else:
                ret[inkey][1].extend([''] * input_channel.length())
            logger.warning(
                'No data found for input key "%s"'
                ', use empty strings/lists instead.' % inkey,
                proc=this.id
            )
        else:
            ret[inkey][1].extend(input_channel.flatten(i))

    return ret

def proc_output(this, value):
    """Parse the output for jobs to easily access it"""
    # ['a:{{i.invar}}', 'b:file:{{i.infile|fn}}']
    output = value
    if isinstance(value, (list, str)):
        outlist = list(filter(None, always_list(value)))
        output = OrderedDiot()
        for out in outlist:
            outparts = safe_split(out, ':')
            lenparts = len(outparts)
            if not outparts[0].isidentifier():
                raise ProcessOutputError(
                    out,
                    'Invalid output idnentifier {!r} in'.format(outparts[0])
                )
            if lenparts < 2:
                raise ProcessOutputError(
                    out,
                    'One of <key>:<type>:<value> missed for process output in'
                )
            if lenparts > 3:
                raise ProcessOutputError(out,
                                         'Too many parts for process output in')
            output[':'.join(outparts[:-1])] = outparts[-1]
    elif not (isinstance(value, OrderedDiot) or
              (isinstance(value, dict) and len(value) == 1)):
        raise ProcessOutputError(
            '%s: Process output type should be one of list/str/OrderedDiot, '
            'or dict with len=1, but get %s' % (this.id, type(value).__name__)
        )

    # output => {'a': '{{i.invar}}', 'b:file': '{{i.infile | fn}}'}
    ret = OrderedDiot()
    for keytype, outdata in output.items():
        if ':' not in keytype:
            keytype += ':' + OUT_VARTYPE[0]
        thekey, thetype = keytype.split(':', 1)

        if (thetype not in (OUT_DIRTYPE +
                            OUT_FILETYPE +
                            OUT_VARTYPE +
                            OUT_STDOUTTYPE +
                            OUT_STDERRTYPE)):
            raise ProcessOutputError(thetype, 'Unknown output type')
        ret[thekey] = (thetype, this.template(outdata, **this.envs))
    return ret

def proc_size(this, value):
    """Deduce the size (# jobs) of the processes from input"""
    _require(this,
             'input',
             strict=False,
             msg='Process size needs input to be initialized.')
    if not this.input:
        return 0
    return len(list(this.input.values())[0][1])

def proc_jobs(this, value):
    """Prepare the jobs"""
    _require(this,
             'size',
             strict=False,
             msg='Jobs need size to be initialized.')
    use_runner(this.runner.runner)
    # try to compute data, since they are not thread-safe
    this.input # pylint: disable=pointless-statement
    this.output # pylint: disable=pointless-statement
    this.script # pylint: disable=pointless-statement
    # have to make sure this is the first time it is imported,
    # as we need the runner to be set correctly for the
    # default value of runner for Job
    from .job import Job
    if this.size == 0:
        logger.warning('No data found for jobs, process will be skipped.',
                       proc=this.id)
        return []
    logger.debug('Constructing jobs ...', proc=this.id)
    return [Job(i, this) for i in range(this.size)]

def proc_lang(this, value):
    """Get the path of the interpreter from lang"""
    return cmdy.which(value).strip()

def proc_script(this, value):
    """Prepare the script"""
    _require(this, 'lang', msg='Process script needs lang to be initialized.')
    _require(this,
             'template',
             msg='Process script needs template to be initialized.')
    if not value:
        logger.warning('No script specified', proc=this.id)
        value = ''

    if value.startswith('file:'):
        tplfile = Path(value[5:])
        if not fs.exists(tplfile):
            raise ProcessScriptError(tplfile, 'No such template file')
        logger.debug("Using template file: %s", tplfile, proc=this.id)
        value = tplfile.read_text()

    value = textwrap.dedent(value)
    if not value.startswith('#!'):
        value = '#!%s\n\n%s' % (this.lang, value)

    return this.template(value.rstrip() + '\n', **this.envs)

def proc_template(this, value):
    """Prepare the template"""
    if callable(value):
        return value
    if not value:
        return template.TemplateLiquid
    return getattr(template, 'Template' + value.capitalize())

def proc_suffix(this, value):
    """Calculate the suffix"""
    _require(this,
             'input',
             msg='Process suffix needs input to be initialized.')
    _require(this,
             'name',
             msg='Process suffix needs name to be initialized.')
    _require(this,
             'output',
             msg='Process suffix needs output to be initialized.')
    _require(this,
             'depends',
             strict=False,
             msg='Process suffix needs depends to be initialized.')
    determines = OrderedDiot()
    # Process belongs to just one program
    determines.argv0 = Path(sys.argv[0]).resolve()
    determines.name = this.name
    if isinstance(this._input, dict):
        determines.input = {}
        for key, val in this._input.items():
            # lambda is not pickable
            # convert others to string to make sure it's pickable. Issue #65
            determines.input[key] = funcsig(val) if callable(val) else str(val)
        determines.input = sorted(determines.input.items())
    else:
        determines.input = str(this._input)
    determines.depends = [proc.name + '#' + proc.suffix
                          for proc in this.depends]
    determines.output = this._output
    return str(uuid.uuid5(uuid.NAMESPACE_URL, str(determines)))[:8]

def proc_workdir(this, value):
    """Get the work directory and try to create it"""
    if not value:
        _require(this,
                 'ppldir',
                 msg='Process workdir needs ppldir to be specified.')
        _require(this,
                 'suffix',
                 msg='Process workdir needs suffix to be initialized.')
        workdir = this.ppldir.joinpath('PyPPL.{}.{}'.format(this.name,
                                                            this.suffix))
    elif '/' not in str(value):
        _require(this,
                 'ppldir',
                 msg='Process workdir needs ppldir to be specified.')
        workdir = this.ppldir / value
    else:
        workdir = Path(value)

    if not workdir.is_dir():
        workdir.mkdir(parents=True)
    return workdir

def proc_channel(this, value):
    """Assign the output of jobs to processes as channel"""
    # make sure it's called after run
    _require(this,
             'runtime_config',
             strict=False,
             msg='Process channel can only be accessed after run.')

    if this.jobs:
        chan = Channel.create([
            tuple(value for _, value in job.output.values())
            for job in this.jobs
        ])
        chan.attach(*this.jobs[0].output.keys())
        return chan
    return value



"""Process for PyPPL"""
from functools import partial
from pathlib import Path
import sys
import traceback
import attr
import toml
from diot import Diot
from attr_property import attr_property, attr_property_class
from varname import varname
from .config import config as default_config
from ._proc import (proc_depends_setter, proc_input, proc_output, proc_script,
                    proc_suffix, proc_lang, proc_name, proc_shortname,
                    proc_jobs, proc_size, proc_runner, proc_template,
                    proc_workdir, proc_input_setter,
                    proc_runtime_config_setter, proc_id_setter,
                    proc_tag_setter, proc_channel, proc_procset,
                    proc_setter_count, proc_config_setter)
from .utils import try_deepcopy, brief_list
from .logger import logger
from .jobmgr import STATES, Jobmgr
from .plugin import pluginmgr, PluginConfig
from .runner import runnermgr
from .pyppl import _register_proc


@attr_property_class
@attr.s(eq=False, slots=True)
class Proc:
    """@API
    Process of a pipeline
    """
    # pylint: disable=too-many-instance-attributes
    # Remember the attr being set, they have the highest priority
    _setcounter = attr.ib(default=attr.Factory(dict), init=False, repr=False)
    # The id of the process
    id = attr_property(
        default=attr.Factory(lambda: varname(caller=2)),
        repr=False,
        setter=proc_id_setter,
        doc="@API\nThe identity of the process")
    # The tag of the process
    tag = attr_property(default=default_config.tag,
                        setter=proc_tag_setter,
                        kw_only=True,
                        repr=False,
                        doc="@API\nThe tag of the process")
    # The description of the process
    desc = attr.ib(default='No description.', kw_only=True, repr=False)
    # The args of the process
    args = attr_property(default=attr.Factory(Diot), kw_only=True, repr=False)
    # caching
    cache = attr_property(
        default=default_config.cache,
        kw_only=True,
        repr=False,
        setter=partial(proc_setter_count, name='cache'),
        doc='@API\nShould we cache the results or read results from cache?')
    # The output channel
    channel = attr_property(default=None,
                            init=False,
                            repr=False,
                            getter=proc_channel)
    # Save the definitions to indicate you when you have 2
    # processes with the same id and tag defined
    _defs = attr.ib(default=attr.Factory(
        lambda: (stack for stack in traceback.format_stack()[:-2]
                 if '<frozen importlib' not in stack)),
                    init=False,
                    repr=False)
    # the dependencies of the process
    depends = attr_property(default=attr.Factory(list),
                            kw_only=True,
                            repr=False,
                            setter=proc_depends_setter,
                            raw='_')
    # dirsig
    dirsig = attr_property(default=default_config.dirsig,
                           kw_only=True,
                           repr=False,
                           setter=partial(proc_setter_count, name='dirsig'))
    # envs used to render templates
    envs = attr.ib(default=try_deepcopy(default_config.envs),
                   kw_only=True,
                   repr=False)
    # how to deal with errors
    errhow = attr_property(default=default_config.errhow,
                           kw_only=True,
                           repr=False,
                           setter=partial(proc_setter_count, name='errhow'))
    # how many times to retry if errored
    errntry = attr_property(default=default_config.errntry,
                            kw_only=True,
                            repr=False,
                            setter=partial(proc_setter_count, name='errntry'))
    # forks
    forks = attr_property(default=default_config.forks,
                          converter=int,
                          converter_runtime=True,
                          kw_only=True,
                          repr=False,
                          setter=partial(proc_setter_count, name='forks'))
    # input of the process
    input = attr_property(default=None,
                          kw_only=True,
                          getter=proc_input,
                          setter=proc_input_setter,
                          repr=False,
                          raw='_')
    # jobs
    jobs = attr_property(default=attr.Factory(list),
                         init=False,
                         repr=False,
                         getter=proc_jobs)
    # language
    lang = attr_property(default=default_config.lang,
                         kw_only=True,
                         repr=False,
                         setter=partial(proc_setter_count, name='lang'),
                         getter=proc_lang)
    # The full name, with procset
    name = attr_property(init=False, getter=proc_name, setter=False, repr=True)
    # Non-cached job indexes
    _ncjobids = attr.ib(default=attr.Factory(list), init=False, repr=False)
    # who depends on me?
    nexts = attr_property(
        init=False,
        # will be automatically deduced from depends, not allowed to set
        setter=False,
        repr=False,
        getter=lambda this, value: [])
    # nthread
    nthread = attr.ib(default=default_config.nthread,
                      converter=int,
                      kw_only=True,
                      repr=False)
    # original job id that I copied from
    origin = attr.ib(default=None, init=False, repr=False)
    # output of the process
    output = attr_property(default='',
                           kw_only=True,
                           getter=proc_output,
                           raw='_',
                           repr=False)
    # pipeline directory
    ppldir = attr_property(default=default_config.ppldir,
                           converter=Path,
                           kw_only=True,
                           repr=False,
                           setter=partial(proc_setter_count, name='ppldir'),
                           converter_runtime=True)
    # name of the procset
    procset = attr_property(setter=False,
                            init=False,
                            repr=False,
                            getter=proc_procset)
    # runner
    runner = attr_property(default=default_config.runner,
                           kw_only=True,
                           repr=False,
                           setter=partial(proc_setter_count, name='runner'),
                           getter=proc_runner,
                           raw='_')
    # runtime configuration, used to override all possible configurations
    runtime_config = attr_property(default=None,
                                   init=False,
                                   repr=False,
                                   setter=proc_runtime_config_setter)
    # script
    script = attr_property(default=None,
                           kw_only=True,
                           repr=False,
                           raw='_',
                           getter=proc_script)
    # Short name without procset
    shortname = attr_property(setter=False,
                              init=False,
                              repr=False,
                              getter=proc_shortname)
    # size of the process
    size = attr_property(init=False,
                         repr=False,
                         getter=proc_size,
                         setter=False)  # deduced from input
    # suffix
    suffix = attr_property(init=False,
                           repr=False,
                           getter=proc_suffix,
                           setter=False)
    # template
    template = attr_property(default=default_config.template,
                             kw_only=True,
                             repr=False,
                             setter=partial(proc_setter_count,
                                            name='template'),
                             getter=proc_template)
    # working directory
    workdir = attr_property(default='',
                            kw_only=True,
                            getter=proc_workdir,
                            repr=False)
    # plugin configs
    config = attr_property(
        # use a callback to make sure values added by plugins to be loaded
        default=PluginConfig(default_config.config),
        setter=proc_config_setter,
        #getter  = config_getter, # default value
        init=True,
        kw_only=True,
        repr=False)
    # properties for plugins
    props = attr.ib(default=Diot(), init=False, repr=False)

    def __attrs_post_init__(self):
        _register_proc(self)
        pluginmgr.hook.proc_init(proc=self)

    def run(self, runtime_config):
        """@API
        Run the process
        @params:
            runtime_config (simpleconf.Config): The runtime configuration
        """
        self.runtime_config = runtime_config
        self.input  # pylint: disable=pointless-statement
        self.output  # pylint: disable=pointless-statement
        self.suffix  # pylint: disable=pointless-statement
        logger.workdir(self.workdir, proc=self.id)
        self._save_settings()
        self._run_jobs()
        self._post_run()

    def _post_run(self):
        jobs = {}

        for job in (self.jobs or []):  # pylint: disable=not-an-iterable
            jobs.setdefault(job.state, []).append(job.index)

        (logger.P_DONE, logger.CACHED)[int(
            len(jobs.get(STATES.DONECACHED, [])) == self.size and self.size > 0
        )]('Jobs [Cached: %s, Succ: %s, B.Fail: %s, S.Fail: %s, R.Fail: %s]',
           len(jobs.get(STATES.DONECACHED, [])),
           len(jobs.get(STATES.DONE, [])),
           len(jobs.get(STATES.BUILTFAILED, [])),
           len(jobs.get(STATES.SUBMITFAILED, [])),
           len(jobs.get(STATES.ENDFAILED, [])),
           proc=self.id)

        logger.debug('Cached: %s',
                     brief_list(jobs.get(STATES.DONECACHED, []), 1),
                     proc=self.id)
        logger.debug('Succeeded: %s',
                     brief_list(jobs.get(STATES.DONE, []), 1),
                     proc=self.id)
        if STATES.BUILTFAILED in jobs:
            logger.error('Building failed: %s',
                         brief_list(jobs[STATES.BUILTFAILED], 1),
                         proc=self.id)
        if STATES.SUBMITFAILED in jobs:
            logger.error('Submission failed: %s',
                         brief_list(jobs[STATES.SUBMITFAILED], 1),
                         proc=self.id)
        if STATES.ENDFAILED in jobs:
            logger.error('Running failed: %s',
                         brief_list(jobs[STATES.ENDFAILED], 1),
                         proc=self.id)

        donejobs = jobs.get(STATES.DONE, []) + jobs.get(STATES.DONECACHED, [])
        if len(donejobs) < self.size and self.errhow != 'ignore':
            pluginmgr.hook.proc_postrun(proc=self, status='failed')
            sys.exit(1)

        logger.debug("Running hook.proc_postrun ...", proc=self.id)
        pluginmgr.hook.proc_postrun(
            proc=self,
            status='cached'
            if jobs and len(jobs.get(STATES.DONECACHED, [])) == self.size else
            'succeeded')

        if self.jobs:
            # pylint: disable=unsupported-delete-operation
            del self.jobs[:]

    def _save_settings(self):
        def stringify(conf):
            """Convert Path to str, even if it is buried deeply"""
            if isinstance(conf, Path):
                return str(conf)
            if isinstance(conf, Proc):
                return conf.name
            if isinstance(conf, dict):
                return {str(key): stringify(val) for key, val in conf.items()}
            if isinstance(conf, list):
                return [stringify(val) for val in conf]
            return str(conf)

        with open(self.workdir / 'proc.settings.toml', 'w') as fsettings:
            toml.dump({key: stringify(val)
                       for key, val in self.__attrs_property_raw__.items()
                       if key not in ('id', 'jobs', 'runtime_config')},
                      fsettings)

    def _run_jobs(self):
        logger.debug('Queue starts ...', proc=self.id)
        # jobs are constructed here, hook.job_init called
        jobmgr = Jobmgr(self.jobs)

        logger.debug("Running hook.proc_prerun ...", proc=self.id)
        ret = pluginmgr.hook.proc_prerun(proc=self)
        # plugins can stop process from running
        # let's put it here since jobs have been init'ed
        # we can then access proc.size, input, etc..
        if ret is False:
            # avoid error in _post_run
            self.jobs = []
            return
        # we need to jobs to be initialized, as during initialization
        # use_runner called, and we only initialize that runner
        runnermgr.hook.runner_init(proc=self)
        jobmgr.start()
        # force calculating channel
        self.channel  # pylint: disable=pointless-statement

    def copy(self, id=None, **kwargs):  # pylint: disable=redefined-builtin, invalid-name
        """@API
        Copy a process to a new one
        Depends and nexts will be copied
        @params:
            id: The id of the new process
            kwargs: Other arguments for constructing a process
        """
        newid = id or varname()
        raw_attrs = {key: try_deepcopy(value)
                     for key, value in self.__attrs_property_raw__.items()
                     if key not in ('input', 'id', 'channel', 'jobs',
                                    'runtime_config', 'depends', 'nexts')}
        # attr.ib not in __attrs_property_raw__
        raw_attrs.update({
            # only keep the keys of input
            'input': (','.join(self._input)
                      if isinstance(self._input, list)
                      else ','.join(self._input.keys())
                      if isinstance(self._input, dict)
                      else self._input),
            'desc': self.desc,
            'envs': try_deepcopy(self.envs),
            'nthread': self.nthread,
        })
        raw_attrs['envs'].update(kwargs.pop('envs', {}))
        raw_attrs.update(kwargs)
        newproc = Proc(newid, **raw_attrs)
        # pylint: disable=attribute-defined-outside-init
        newproc.origin = self.origin or self.id
        return newproc

    def add_config(self, name, default=None, converter=None, runtime='update'):
        """@API
        Add a plugin configuration
        @params:
            name (str): The name of the plugin configuration
            default (any): The default value
            converter (callable): The converter function for the value
            runtime (str): How should we deal with it while \
                runtime_config is passed and its setcounter > 1
                - override: Override the value
                - update: Update the value if it's a dict otherwise override its
                - ignore: Ignore the value from runtime_config
        """
        self.config.add(name, default, converter, runtime)



"""
Plugin system for PyPPL
@variables:
    PMNAME (str): The name of the plugin manager
    hookimpl (pluggy.HookimplMarker): Used to mark the implementation of hooks
    hookspec (pluggy.HookspecMarker): Used to mark the hooks
"""
# pylint: disable=unused-argument
import sys
import types
import pluggy
from diot import Diot
from .exception import PluginNoSuchPlugin

PMNAME = "pyppl"

# pylint: disable=invalid-name
hookimpl = pluggy.HookimplMarker(PMNAME)
hookspec = pluggy.HookspecMarker(PMNAME)


@hookspec(
    warn_on_impl=DeprecationWarning("This hook is deprecated. "
                                    "Please initialize configs inside "
                                    "your plugin")
)
def setup(config):
    """@API
    PLUGIN API
    Add default configs
    Note that this is not running for runtime plugins, which are regiested later
    after the module is loaded.
    @params:
        config (Config): The default configurations
    """

@hookspec
def proc_init(proc):
    """@API
    PLUGIN API
    Right after a Proc being initialized
    @params:
        proc (Proc): The Proc instance
    """


@hookspec(firstresult=True)
def proc_prerun(proc):
    """@API
    PLUGIN API
    Before a process starts
    If False returned, process will not start
    The value returned by the first plugin will be used, which means
    once a plugin stops process from running, others cannot resume it.
    @params:
        proc (Proc): The Proc instance
    """

@hookspec
def proc_postrun(proc, status):
    """@API
    PLUGIN API
    After a process has done
    @params:
        proc (Proc): The Proc instance
        status (str): succeeded/failed/cached
    """


@hookspec
def pyppl_init(ppl):
    """@API
    PLUGIN API
    Right after a pipeline initiates
    @params:
        ppl (PyPPL): The PyPPL instance
    """


@hookspec(firstresult=True)
def pyppl_prerun(ppl):
    """@API
    PLUGIN API
    Before pipeline starts to run
    If False returned, the pipeline will not run
    The value returned by the first plugin will be used, which means
    once a plugin stops process from running, others cannot resume it.
    @params:
        ppl (PyPPL): The PyPPL instance
    """


@hookspec(firstresult=True)
def pyppl_postrun(ppl):
    """@API
    PLUGIN API
    After the pipeline is done
    If you want to do something after a process fails
    Use proc_postrun(proc = proc, status = 'failed') instead.
    Return False will stop exception being raised
    @params:
        ppl (PyPPL): The PyPPL instance
    """


@hookspec
def job_init(job):
    """@API
    PLUGIN API
    Right after job initiates
    @params:
        job (Job): The Job instance
    """


@hookspec
def job_succeeded(job):
    """@API
    PLUGIN API
    Tell if job is successfully done or not
    One can add not rigorous check. By default, only
    if returncode is 0 checked.
    return False to tell if job is failed otherwise
    use the default status or results from other plugins
    @params:
        job (Job): The Job instance
    """


@hookspec
def job_prebuild(job):
    """@API
    PLUGIN API
    Before a job starts to build
    @params:
        job (Job): The Job instance
    """


@hookspec
def job_build(job, status):
    """@API
    PLUGIN API
    After a job is being built
    @params:
        job (Job): The Job instance
        status (str): The status of the job building
            - True: The job is successfully built
            - False: The job is failed to build
            - cached: The job is cached
    """


@hookspec
def job_submit(job, status):
    """@API
    PLUGIN API
    After a job is being submitted
    @params:
        job (Job): The Job instance
        status (str): The status of the job submission
            - 'succeeded': The job is successfully submitted
            - 'failed': The job is failed to submit
            - 'running': The job is already running
    """


@hookspec
def job_poll(job, status):
    """@API
    PLUGIN API
    Poll the status of a job
    @params:
        job (Job): The Job instance
        status (str): The status of the job
            - 'running': The job is still running
            - 'done': Polling is done, rcfile is generated
    """


@hookspec
def job_kill(job, status):
    """@API
    PLUGIN API
    After a job is being killed
    @params:
        job (Job): The Job instance
        status (str): The status of the job killing
            - 'succeeded': The job is successfully killed
            - 'failed': The job is failed to kill
    """


@hookspec
def job_done(job, status):
    """@API
    PLUGIN API
    After a job is done
    @params:
        job (Job): The Job instance
        status (str): The status of the job
            - succeeded: The job is successfully done
            - failed: The job is failed
            - cached: The job is cached
    """


@hookspec
def logger_init(logger):
    """@API
    PLUGIN API
    Initiate logger, most manipulate levels
    @params:
        logger (Logger): The Logger instance
    """


@hookspec
def cli_addcmd(commands):
    """@API
    PLUGIN API
    Add command and options to CLI
    @params:
        commands (Commands): The Commands instance
    """


@hookspec
def cli_execcmd(command, opts):
    """@API
    PLUGIN API
    Execute the command being added to CLI
    @params:
        command (str): The command
        opts (dict): The options
    """


pluginmgr = pluggy.PluginManager(PMNAME)
pluginmgr.add_hookspecs(sys.modules[__name__])


def _get_plugin(name):
    """
    Try to find the plugin by name with/without prefix.
    If the plugin is not found, try to treat it as a module and import it
    """
    if isinstance(name, str):
        for plugin in pluginmgr.get_plugins():
            plname = pluginmgr.get_name(plugin)
            if plname.isdigit():
                plname = plugin.__class__.__name__
            if plname in ('pyppl-' + name, 'pyppl_' + name,
                          'PyPPL' + name.capitalize()):
                return plugin
        try:
            __import__(name)
        except ImportError as exc:
            raise PluginNoSuchPlugin(name) from exc
    return name


def disable_plugin(plugin):
    """@API
    Try to disable a plugin
    @params:
        plugin (any): A plugin or the name of a plugin
    """
    plugin = _get_plugin(plugin)
    if pluginmgr.is_registered(plugin):
        pluginmgr.unregister(plugin)


def config_plugins(*plugins):
    """@API
    Parse configurations for plugins and enable/disable plugins accordingly.
    @params:
        *plugins ([any]): The plugins
            plugins with 'no:' will be disabled.
    """
    for plugin in plugins:
        if isinstance(plugin, str) and plugin[:3] == 'no:':
            try:
                disable_plugin(_get_plugin(plugin[3:]))
            except PluginNoSuchPlugin:
                pass
        else:
            plugin = _get_plugin(plugin)
            if not pluginmgr.is_registered(plugin):
                if isinstance(plugin, types.ModuleType):
                    pluginmgr.register(plugin)
                else:
                    pluginmgr.register(plugin, name=plugin.__class__.__name__)


class PluginConfig(Diot):
    """@API
    Plugin configuration for Proc/Job"""
    def __init__(self, *args, **kwargs):
        """@API
        Construct for PluginConfig
        @params:
            pconfig (dict): the default plugin configuration
        """
        self.__dict__['_meta'] = Diot(raw={},
                                      converter={},
                                      setcounter={},
                                      updates={})
        super().__init__(*args, **kwargs)
        # reset all counters
        for key in self._meta.setcounter:
            self._meta.setcounter[key] = 0

    def add(self, name, default=None, converter=None, update='update'):
        """@API
        Add a config item
        @params:
            name (str): The name of the config item.
            default (any): The default value
            converter (callable): The converter to convert the value
                whenever the value is set.
            update (str): With setcounter > 1,
                should we update the value or ignore it in .update()?
                - You can set plugin_config_check_update to False with
                    .update to disable this
                - Could be ignore (don't update),
                    replace (replace the whole value, even it is a
                  dictionary) or update
                    (replace the non-dict value and update dictionary values)
        """
        self._meta.converter[name] = converter
        self._meta.updates[name] = update
        # setitem will run the converter
        # if value exists, use it.
        self[name] = self._meta.raw.get(name, default)
        # reset counter
        self._meta.setcounter[name] = 0
        return self

    def update(self, *args, **kwargs):
        """@API
        Update the configuration
        Depends on `update` argument while the configuration is added
        @params:
            pconfig (dict): the configuration to update from
        """
        # if we should check using meta update information or
        # just update directly
        plugin_config_check_update = kwargs.pop("plugin_config_check_update",
                                                True)
        dict_to_update = dict(*args, **kwargs)

        for key, value in dict_to_update.items():

            if (plugin_config_check_update and
                    self._meta.updates.get(key, 'update') == 'ignore' and
                    self._meta.setcounter.get(key, 0) > 0):
                continue

            if (plugin_config_check_update and
                    self._meta.updates.get(key, 'update') == 'update' and
                    isinstance(value, dict)):
                if key in self and isinstance(self[key], dict):
                    self[key].update(value)
                else:
                    self[key] = value
            else:
                self[key] = value

    def __setitem__(self, name, value):
        if name[:6] != '_diot_':
            self._meta.setcounter[
                name
            ] = self._meta.setcounter.setdefault(name, 0) + 1
            self._meta.raw[name] = value
            if self._meta.converter.get(name):
                value = self._meta.converter[name](value)
        super().__setitem__(name, value)



"""Job manager for PyPPL
@variables:
    STATES (dict): Possible states for the job
    PBAR_MARKS (dict): The marks on progress bar for different states
    PBAR_LEVEL (dict): The levels for different states
"""
import sys
from time import sleep
from threading import Lock
from queue import Queue
from diot import Diot
from .utils import StateMachine, PQueue, ThreadPool, log_msg_len
from .plugin import pluginmgr
from .logger import logger
from .exception import JobBuildingError, JobFailError
from .runner import poll_interval

STATES = Diot(
    INIT='00_init',
    BUILDING='99_building',
    BUILT='97_built',
    BUILTFAILED='98_builtfailed',
    SUBMITTING='89_submitting',
    SUBMITFAILED='87_submitfailed',
    RUNNING='78_running',
    DONE='67_done',
    RETRYING='80_retrying',
    DONECACHED='66_donecached',
    DONEFAILED='68_donefailed',
    ENDFAILED='69_endfailed',
    KILLING='59_killing',
    KILLED='57_killed',
    KILLFAILED='58_killfailed',
)
PBAR_MARKS = {
    STATES.INIT: ' ',
    STATES.BUILDING: '~',
    STATES.BUILT: '-',
    STATES.BUILTFAILED: '!',
    STATES.SUBMITTING: '>',
    STATES.SUBMITFAILED: '$',
    STATES.RUNNING: '>',
    STATES.RETRYING: '-',
    STATES.DONE: '=',
    STATES.DONECACHED: 'z',
    STATES.DONEFAILED: 'x',
    STATES.ENDFAILED: 'X',
    STATES.KILLING: '<',
    STATES.KILLED: '*',
    STATES.KILLFAILED: '*',
}
PBAR_LEVEL = {
    STATES.INIT: 'BLDING',
    STATES.BUILDING: 'BLDING',
    STATES.BUILT: 'BLDING',
    STATES.BUILTFAILED: 'BLDING',
    STATES.SUBMITTING: 'SBMTING',
    STATES.SUBMITFAILED: 'SBMTING',
    STATES.RUNNING: 'RUNNING',
    STATES.RETRYING: 'RTRYING',
    STATES.DONE: 'JOBDONE',
    STATES.DONECACHED: 'JOBDONE',
    STATES.DONEFAILED: 'JOBDONE',
    STATES.ENDFAILED: 'JOBDONE',
    STATES.KILLING: 'KILLING',
    STATES.KILLED: 'KILLING',
    STATES.KILLFAILED: 'KILLING',
}


class Jobmgr:
    """@API
    Job manager"""

    __slots__ = ('jobs', 'proc', 'stop', 'queue', 'nslots', 'lock', 'barsize')

    def __init__(self, jobs):
        """@API
        Job manager constructor
        @params:
            jobs (list): All jobs of a process
        """
        if not jobs:  # no jobs
            return

        self.lock = Lock()

        machine = StateMachine(model=jobs,
                               states=list(STATES.values()),
                               initial=STATES.INIT,
                               send_event=True,
                               after_state_change=self.progressbar)

        self.jobs = jobs
        self.proc = jobs[0].proc
        self.stop = False

        self.queue = PQueue(batch_len=len(jobs))
        self.nslots = min(self.queue.batch_len, int(self.proc.nthread))

        self.barsize = log_msg_len(procid=self.proc.id,
                                   joblen=len(jobs))

        for job in jobs:
            self.queue.put(job.index)

        # start building
        machine.add_transition(
            trigger='trigger_build',
            source=STATES.INIT,
            dest={
                'cached': STATES.DONECACHED,
                True: STATES.BUILT,
                False: STATES.BUILTFAILED
            },
            depends_on='build',
            before=lambda event: setattr(event.model,
                                         'state',
                                         STATES.BUILDING),
            after=self._post_event
        )

        # do the real submit
        machine.add_transition(
            trigger='trigger_submit',
            source=[STATES.BUILT, STATES.RETRYING],
            dest={
                True: STATES.RUNNING,
                False: STATES.SUBMITFAILED
            },
            depends_on='submit',
            after=self._post_event
        )

        # try to retry if submission failed or job itself failed
        machine.add_transition(
            trigger='trigger_retry',
            source=[STATES.SUBMITFAILED, STATES.DONEFAILED],
            dest={
                True: STATES.RETRYING,  # ready to re-submit
                'ignored': STATES.DONE,
                False: STATES.ENDFAILED
            },
            depends_on='retry',
            before=lambda event: sleep(.5),
            after=self._post_event
        )

        # do the poll for the results
        machine.add_transition(
            trigger='trigger_poll',
            source=STATES.RUNNING,
            dest={
                'running': STATES.RUNNING,
                True: STATES.DONE,
                False: STATES.DONEFAILED
            },
            depends_on='poll',
            after=self._post_event
        )

        # killed/failed to kill
        machine.add_transition(
            trigger='trigger_kill',
            # STATES.KILLING not guareteed, as this will be
            # running in a separate queue
            source='*',
            dest={
                True: STATES.KILLED,
                False: STATES.KILLFAILED
            },
            before=lambda event: setattr(event.model, 'state', STATES.KILLING),
            depends_on='kill'
        )

    def start(self):
        """@API
        Start the queue.
        """
        # no jobs
        try:
            self.jobs
        except AttributeError:
            return

        pool = ThreadPool(self.nslots, initializer=self.worker)
        pool.join(cleanup=self.cleanup)
        self.progressbar(Diot(model=self.jobs[-1]))

    def _get_jobs_by_states(self, *states):
        return [job for job in self.jobs if job.state in states]

    def _distribute_jobs_to_pbar(self):
        joblen = len(self.jobs)
        index_bjobs = []
        if joblen <= self.barsize:
            div, mod = divmod(self.barsize, joblen)
            for j in range(joblen):
                step = div + 1 if j < mod else div
                for _ in range(step):
                    index_bjobs.append([j])
        else:
            jobx = 0
            div, mod = divmod(joblen, self.barsize)
            for i in range(self.barsize):
                step = div + 1 if i < mod else div
                index_bjobs.append([jobx + jobstep for jobstep in range(step)])
                jobx += step
        return index_bjobs

    def progressbar(self, event):
        """@API
        Generate the progress bar.
        @params:
            event (StateMachine event): The event including job as model.
        """
        job = event.model
        index_bjobs = self._distribute_jobs_to_pbar()

        # get all states in this moment
        #with self.lock:
        states = [job.state for job in self.jobs]
        ncompleted = nrunning = 0
        for state in states:
            ncompleted += int(state in (STATES.DONE, STATES.DONECACHED,
                                        STATES.ENDFAILED))
            nrunning += int(state in (STATES.RUNNING, STATES.SUBMITTING))

        pbar = '['
        pbar += ''.join(PBAR_MARKS[states[job.index]]
                        if job.index in indexes
                        else PBAR_MARKS[max(states[index]
                                            for index in indexes)]
                        for indexes in index_bjobs)
        pbar += '] Done: {:5.1f}% | Running: {}'.format(
            100.0 * float(ncompleted) / float(len(self.jobs)),
            str(nrunning).ljust(len(str(self.proc.forks))))

        job.logger(pbar,
                   pbar=True,
                   done=ncompleted == len(self.jobs),
                   level=PBAR_LEVEL[states[job.index]])

    def cleanup(self, ex=None):
        """@API
        Cleanup the pipeline when
        - Ctrl-C hit
        - error encountered and `proc.errhow` = 'terminate'
        @params:
            ex (Exception): The exception raised by workers
        """
        self.stop = True
        message = None
        if isinstance(ex, JobBuildingError):
            message = ('Job building failed, quitting pipeline '
                       '(Ctrl-C to skip killing jobs) ...')
        elif isinstance(ex, JobFailError):
            message = ('Error encountered (errhow = halt), quitting pipeline '
                       '(Ctrl-C to skip killing jobs) ...')
        elif isinstance(ex, KeyboardInterrupt):
            message = ('[Ctrl-C] detected, quitting pipeline '
                       '(Ctrl-C again to skip killing jobs) ...')

        if message:
            logger.warning(message, proc=self.proc.id)

        # kill running jobs
        with self.lock:

            # failed_jobs =	self._get_jobs_by_states(STATES.ENDFAILED) or \
            # 				self._get_jobs_by_states(STATES.DONEFAILED) or \
            # 				self._get_jobs_by_states(STATES.SUBMITFAILED) or \
            # 				self._get_jobs_by_states(STATES.BUILTFAILED)

            running_jobs = self._get_jobs_by_states(
                # all possible states to go to next steps
                STATES.BUILT,
                STATES.SUBMITTING,
                STATES.RUNNING,
                STATES.RETRYING,
                STATES.DONEFAILED,
            )
            killq = Queue()
            for rjob in running_jobs:
                killq.put(rjob)

            try:
                ThreadPool(min(len(running_jobs), self.proc.nthread),
                           initializer=self.kill_worker,
                           initargs=killq).join()
            # not enough resources to start the queue
            except (RuntimeError, BlockingIOError):  # pragma: no cover
                logger.warning('Not enough resources '
                               '(probably jobs with too many threads) '
                               'to kill jobs, skip.')

            pluginmgr.hook.proc_postrun(proc=self.proc, status='failed')
            #random.choice(failed_jobs or running_jobs or self.jobs)
            # .showError(len(failed_jobs))

            if isinstance(ex, Exception) and not isinstance(
                    ex, (JobFailError, JobBuildingError, KeyboardInterrupt)):
                raise ex from None

            sys.exit(1)

    @classmethod
    def kill_worker(cls, killq):
        """@API
        The killing worker to kill the jobs"""
        while not killq.empty():
            job = killq.get()
            job.trigger_kill()
            killq.task_done()

    def _post_event(self, event):
        job = event.model
        if job.state == STATES.DONECACHED:
            job.done(cached=True)
        elif job.state in (STATES.BUILT, STATES.RETRYING):
            self.queue.put_next(job.index, event.kwargs['batch'])
        elif job.state == STATES.BUILTFAILED:
            raise JobBuildingError()
        elif job.state == STATES.RUNNING:
            self.queue.put_next(job.index, event.kwargs['batch'])
        elif job.state in (STATES.SUBMITFAILED, STATES.DONEFAILED):
            job.trigger_retry(batch=event.kwargs['batch'])
        elif job.state == STATES.DONE:
            job.done(status=True)
        elif job.state == STATES.ENDFAILED:
            if self.proc.errhow == 'halt':
                raise JobFailError()
            # else: endfailed but ignored
            job.done(status=False)

    def worker(self):
        """@API
        The worker to build, submit and poll the jobs"""
        while not self.queue.empty() and not self.stop:
            index, batch = self.queue.get()
            job = self.jobs[index]
            # make sure quit me for kill_worker
            # and don't use a GIL, as it blocks regular jobs
            if self.stop:  # pragma: no cover
                break

            if job.state == STATES.INIT:
                job.trigger_build(batch=batch)
            elif job.state in (STATES.BUILT, STATES.RETRYING):
                with self.lock:
                    if len(
                            self._get_jobs_by_states(
                                STATES.RUNNING,
                                STATES.SUBMITTING)) < self.proc.forks:
                        try:
                            job.trigger_submit(batch=batch)
                        except (RuntimeError,
                                BlockingIOError) as rbex:  # pragma: no cover
                            if ("can't start new thread" not in str(rbex)
                                    and "Resource temporarily unavailable"
                                    not in str(rbex)):
                                raise
                            # wait longer for heavy multithreading jobs
                            sleep(5)
                # if we successfully submitted
                if job.state in (STATES.BUILT, STATES.RETRYING):
                    sleep(1)
                    # put the job back to the queue
                    self.queue.put_next(index, batch)
            elif job.state == STATES.RUNNING:
                # have to be longer than ThreadPool.join's interval
                sleep(poll_interval())
                job.trigger_poll(batch=batch)
            elif job.state == STATES.KILLING:  # pragma: no cover
                break
            #else: # endfailed but ignored, after retry
            #	pass

            self.queue.task_done()



"""
The procset for a set of procs
"""
import inspect
from types import GeneratorType
from fnmatch import fnmatch, filter as fnfilter
from diot import Diot, OrderedDiot
from varname import varname


class Proxy(list):
    """
    A proxy class extended from list to enable dot access
    to all members and set attributes for all members.
    """
    def __getattr__(self, item):
        try:
            # Get the attributes of list
            return getattr(super(Proxy, self), item)
        except AttributeError:
            return self.__class__(getattr(proxy, item) for proxy in self)

    def __setattr__(self, name, value):
        # We are unable to setattr of existing attribute of list
        #if hasattr(super(Proxy, self), name):
        #	super(Proxy, self).__setattr__(name, value)
        if isinstance(value, Values):
            for i, val in enumerate(value):
                setattr(self[i], name, val)
        else:
            for proxy in self:
                setattr(proxy, name, value)

    def __getitem__(self, item):
        if isinstance(item, int):
            return super(Proxy, self).__getitem__(item)
        if isinstance(item, slice):
            return self.__class__(super(Proxy, self).__getitem__(item))
        return self.__getattr__(item)

    def __setitem__(self, item, value):
        if isinstance(item, (int, slice)):
            return super(Proxy, self).__setitem__(item, value)
        return self.__setattr__(item, value)

    def add(self, anything):
        """
        Add elements to the list.
        @params:
            `anything`: anything that is to be added.
                If it is a Proxy, element will be added individually
                Otherwise the whole `anything` will be added as one element.
        """
        if not anything:
            return
        if isinstance(anything, Proxy):
            for thing in anything:
                self.add(thing)
        elif anything not in self:
            self.append(anything)


class Values(Proxy):
    """A Proxy class but element can be passed one by one"""
    def __init__(self, *args, **kwargs):
        super().__init__(args, **kwargs)


class PSProxy:
    """A Proxy for procset"""
    def __init__(self, procset, path=None):
        self.__dict__['procset'] = procset
        self.__dict__['path'] = path or []

    def _delegated_attrs(self, attr_to_set):
        path_to_check = '.'.join(self.path + [attr_to_set])
        for dele_name in self.procset.delegates.keys():
            if fnmatch(path_to_check, dele_name):
                procs = self.procset.delegated(dele_name)
                break
        else:
            procs = Proxy(self.procset.procs.values())
        for pat in self.path:
            procs = getattr(procs, pat)
        return procs

    def __getattr__(self, item):
        self.path.append(item)
        return self

    def __setattr__(self, name, value):
        attrs = self._delegated_attrs(name)
        if isinstance(value, Values):
            for i, val in enumerate(value):
                setattr(attrs[i], name, val)
        else:
            for attr in attrs:
                setattr(attr, name, value)


class ProcSet:
    """@API
    The ProcSet for a set of processes
    """

    # pylint: disable=redefined-builtin,invalid-name
    def __init__(self, *procs, id=None, tag=None, copy=True, depends=True):
        """@API
        Constructor
        @params:
            *procs (Proc) : the set of processes
            **kwargs: Other arguments to instantiate a `ProcSet`
                depends (bool): Whether auto deduce depends.
                    Default: `True`
                id (str): The id of the procset.
                    Default: `None` (the variable name)
                tag (str): The tag of the processes. Default: `None`
                copy (bool): Whether copy the processes or just use them.
                    Default: `True`
        """

        self.__dict__['id'] = id or varname()
        self.__dict__['tag'] = tag
        self.__dict__['starts'] = Proxy()
        self.__dict__['ends'] = Proxy()
        self.__dict__['delegates'] = OrderedDiot(diot_nest=False)
        self.__dict__['procs'] = OrderedDiot(diot_nest=False)
        self.__dict__['modules'] = Diot(diot_nest=False)
        # save initial states before a module is called
        # states will be resumed before each module is called
        self.__dict__['initials'] = Diot(diot_nest=False)

        prevproc = None
        for proc in procs:
            assert hasattr(proc, 'id') and hasattr(proc, 'tag'), (
                'Argument has to be a Proc object: %r.' % proc
            )
            if copy:
                self.procs[proc.id] = proc.copy(
                    proc.id,
                    tag=(self.tag or proc.tag.split('@', 1)[0]) + '@' +
                    self.id)
            else:
                self.procs[proc.id] = proc
                proc.tag = (self.tag
                            or proc.tag.split('@', 1)[0]) + '@' + self.id

            if depends and prevproc is None:
                self.starts.add(self[proc.id])

            if depends and prevproc:
                self.procs[proc.id].depends = prevproc

            prevproc = self.procs[proc.id]

        if depends and prevproc:
            self.ends.add(prevproc)

        self.delegate('input', 'starts')
        self.delegate('depends', 'starts')
        # export functions moved to plugins
        # self.delegate('ex*', 'ends')

    def delegate(self, attr, *procs):
        """@API
        Delegate process attributes to procset.
        @params:
            *procs (str|Proc): The first argument is the name of the attributes.
                - The rest of them should be `Proc`s or `Proc` selectors.
        """
        procs = list(procs)
        self.delegates[attr] = procs

    def delegated(self, name):
        """@API
        Get the detegated processes by specific attribute name
        @params:
            name (str): the attribute name to query
        @returns:
            (Proxy): The set of processes
        """
        if name not in self.delegates:
            return None
        return self[self.delegates[name]]

    def restore_states(self):
        """@API
        Restore the initial state of a procset
        """
        if not self.initials:  # extract the inital states
            self.initials.starts = self.starts[:]
            self.initials.ends = self.ends[:]
            self.initials.depends = {
                pid: proc.depends
                for pid, proc in self.procs.items()
            }
        else:
            self.__dict__['starts'] = self.initials.starts[:]
            self.__dict__['ends'] = self.initials.ends[:]
            for pid, depends in self.initials.items():
                self.procs[pid] = depends

    def module(self, name):
        """@API
        A decorator used to define a module.
        @params:
            name (callable|str): The function to be decorated or
                the name of the module.
        @returns:
            (callable): The decorator
        """
        if callable(name):
            funcname = name.__name__
            if funcname.startswith(self.id + '_'):
                funcname = funcname[(len(self.id) + 1):]
            return self.module(funcname)(name)

        def decorator(func):
            signature = inspect.signature(func)
            defaults = {
                key: val.default
                for key, val in signature.parameters.items()
                if val.default is not inspect.Parameter.empty
            }

            def modfun(*args, **kwargs):
                if kwargs.get('restore', defaults.get('restore', True)):
                    self.restore_states()
                func(self, *args, **kwargs)

            self.modules[name] = modfun
            return self.modules[name]

        return decorator

    # pylint: disable=arguments-differ,redefined-builtin,unused-argument,invalid-name
    def copy(self, id=None, tag=None, depends=True):
        """@API
        Like `proc`'s `copy` function, copy a procset.
        Each processes will be copied.
        @params:
            id (str): Use a different id if you don't
                want to use the variant name
            tag (str): The new tag of all copied processes
            depends (bool): Whether to copy the dependencies or not.
                Default: True
                - dependences for processes in starts will not be copied
        @returns:
            (ProcSet): The new procset
        """
        id = id or varname()
        ret = self.__class__(*self.procs.values(),
                             id=id,
                             tag=tag,
                             copy=True,
                             depends=False)

        if depends:
            for proc in ret.procs.values():
                proc.depends = [
                    ret.procs[dep.id] if dep is self.procs[dep.id] else dep
                    for dep in self.procs[proc.id].depends
                ]

            ret.starts.add(Proxy(ret.procs[proc.id] for proc in self.starts))
            ret.ends.add(Proxy(ret.procs[proc.id] for proc in self.ends))

        return ret

    def __repr__(self):
        return 'ProcSet(name="%s%s")' % (self.id,
                                         '.' + self.tag if self.tag else '')

    def __setattr__(self, item, value):
        if item in ('starts', 'ends'):
            self.__dict__[item] = self[value]
        elif item in ('id', 'tag'):
            self.__dict__[item] = value
        else:
            PSProxy(procset=self).__setattr__(item, value)

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]
        if item in self.procs:
            return self.procs[item]
        return PSProxy(procset=self, path=[item])

    def __getitem__(self, item, _ignore_default=True):  # pylint:disable=too-many-return-statements
        """@API:
        Process selector, always return Proxy object
        @params:
            item (any): The process selector.
        @returns:
            (Proxy): The processes match the item."""
        if item in ('starts', 'ends'):
            return self.__getattr__(item)
        if hasattr(item, 'id') and hasattr(
                item, 'tag') and not isinstance(item, ProcSet):
            return Proxy([self.procs[item.id]])
        if isinstance(item, slice):
            return Proxy(
                self.__getattr__(procid)
                for procid in list(self.procs.keys())[item])
        if isinstance(item, int):
            return self[list(self.procs.keys())[item]]
        if isinstance(item, (tuple, list, GeneratorType)):
            ret = Proxy()
            ret.add(Proxy(it for itm in item for it in self[itm]))
            return ret
        if item in self.procs:
            return Proxy([self.procs[item]])
        if ',' in item:
            return self[(it.strip() for it in item.split(','))]

        return self[fnfilter(self.procs.keys(), item)]



"""List work directories under <wdir>."""
import sys
from os import path
from glob import glob
from datetime import date, timedelta, datetime
from ..plugin import hookimpl
from ..logger import logger


def show_proc(proc, mtime, fail):
    """Show the status of a process"""
    (logger.error if fail else logger.info)('{} {}: {}'.format(
        'x' if fail else 'v', path.basename(proc),
        mtime.strftime("%Y-%m-%d %H:%M:%S")))


def list_procs(procs, wdir):
    """List all processes"""
    logger.workdir(wdir)
    procname = None
    total = 0
    for proc, mtime, fail in procs:
        total += 1
        pname = '.'.join(path.basename(proc).split('.')[1:3])
        if pname != procname:
            procname = pname
            logger.process(pname)
        show_proc(proc, mtime, fail)
    logger.info('TOTAL: {}'.format(total))


def proc_mtime(proc):
    """Get the modification time of the process"""
    setfile = path.join(proc, 'proc.settings.toml')
    if not path.isfile(setfile):
        return datetime.fromtimestamp(0)
    return datetime.fromtimestamp(path.getmtime(setfile))


def proc_failed(proc):
    """Tell if the process is failed or not"""
    for job in glob(path.join(proc, '*')):
        if not path.basename(job).isdigit():
            continue
        rcfile = path.join(job, 'job.rc')
        if not path.exists(rcfile):
            return True
        with open(rcfile) as frc:
            rc = frc.read().strip()
        if rc != '0':
            return True
    return False


def get_procs(opts):  # pylint: disable=too-many-branches
    """Get the processes"""
    if opts.proc:
        pattern = path.join(opts.wdir, 'PyPPL.{}.*'.format(opts.proc))
    else:
        pattern = path.join(opts.wdir, 'PyPPL.*')

    procs = sorted(glob(pattern))
    if not procs:
        logger.warning('No query processes found in workdir: {}.'.format(
            opts.wdir))
        sys.exit(1)

    before = None
    if opts.ago:
        before = date.today() - timedelta(days=opts.ago)
        before = datetime(before.year, before.month, before.day)
    elif opts.before:
        year = date.today().year
        if opts.before.count('/'):
            parts = opts.before.split('/')
            if len(parts) == 2:
                month, day = parts
            else:
                month, day, year = parts
        else:
            parts = opts.before.split('-')
            if len(parts) == 2:
                month, day = parts
            else:
                year, month, day = parts
        before = datetime(int(year), int(month), int(day))

    for proc in procs:
        mtime = proc_mtime(proc)
        if before and mtime >= before:
            continue

        if opts.nocheck:
            yield (proc, mtime, False)
        else:
            fail = proc_failed(proc)
            if opts.error and not fail:
                continue
            yield (proc, mtime, fail)


@hookimpl
def cli_addcmd(commands):
    """Add command"""
    commands.list = __doc__
    commands.list._hbald = False
    commands.list.proc.desc = 'The process name to show or to compare.'
    commands.list.ago.type = 'int'
    commands.list.ago.desc = ('Work directories to be removed when '
                              'modified N days ago.')
    commands.list.before.desc = [
        'Before when the work directories to be listed.',
        'Supported format: m/d, m-d, m/d/y and y-m-d'
    ]
    commands.list.nocheck = False
    commands.list.nocheck.desc = 'Don`t check failure of processes.'
    commands.list.error = False
    commands.list.error.desc = ('Remove directories if any job failed '
                                'or do error check when listing them.')
    commands.list.wdir = './workdir'
    commands.list.wdir.desc = ('The <ppldir> containing process '
                               'work directories.')


@hookimpl
def cli_execcmd(command, opts):
    """Run the command"""
    if command == 'list':
        list_procs(get_procs(opts), opts.wdir)



"""A thread-safe file operation package for python"""
import os
import shutil
import tempfile
from hashlib import md5
from contextlib import contextmanager
from threading import Lock
from filelock import FileLock
import cmdy

MULTILOCK = Lock()
TMPDIR = tempfile.gettempdir()
#TMPDIR   = os.path.join(tempfile.gettempdir(), 'fsutil.locks')
## not threading-safe!!
# if not os.path.exists(TMPDIR): # pragma: no cover
# 	os.makedirs(TMPDIR)


class TargetExistsError(OSError):
    """Raise when target exists and not able to overwrite"""


class TargetNotExistsError(OSError):
    """Raise when target does not exist and not able to ignore"""


def _get_lock_file(path):
    basename = md5(str(path).encode()).hexdigest()
    return os.path.join(TMPDIR, basename + '.lock')


@contextmanager
def lock(*files, resolve=False):
    """Lock file context"""
    files = [
        os.path.realpath(path) if resolve else os.path.abspath(path)
        for path in files
    ]
    locks = [FileLock(_get_lock_file(path)) for path in files]
    if len(files) == 1:
        with locks[0]:
            yield locks[0]
    else:
        with MULTILOCK:
            _ = [lock.acquire() for lock in locks]
            yield locks
        _ = [lock.release() for lock in locks]


def autolock(func, *filekws, resolve=False):
    """Regiester function for auto locking operating files"""
    def realfunc(*args, **kwargs):
        files = [path for i, path in enumerate(args) if i in filekws]
        files.extend([path for kw, path in kwargs.items() if kw in filekws])
        files = set(files or [args[0]])
        with lock(*files, resolve=resolve) as locks:
            func.lock = locks
            return func(*args, **kwargs)

    return realfunc


exists = os.path.exists  # pylint: disable=invalid-name
isfile = os.path.isfile  # pylint: disable=invalid-name
isdir = os.path.isdir  # pylint: disable=invalid-name
islink = os.path.islink  # pylint: disable=invalid-name


# haven't figured out a way to mimic this
def _remove_busy_dir(path):  # pragma: no cover
    """Try to remove directory with files being ocupied by open process"""
    for root, _, files in os.walk(path):
        for fname in files:
            if not fname.startswith('.'):
                continue
            fpath = os.path.join(root, fname)
            lsof = cmdy.lsof(fpath, _raise=False)  # pylint: disable=no-member
            if lsof.rc != 0:
                continue
            lsof = lsof.splitlines()
            if len(lsof) < 2:
                continue
            pid = lsof[-1].split()[1]
            cmdy.kill({'9': pid})  # pylint: disable=no-member
    shutil.rmtree(path)


def remove(path, ignore_nonexist=True):
    """Remove anything. If ignore_nonexist is False and path does not exists,
    a TargetNotExistsError raises."""
    if os.path.islink(path):
        os.remove(path)
    if os.path.isdir(path):
        try:
            shutil.rmtree(path)
        except OSError as ex:  # pragma: no cover
            if 'busy' not in str(ex):
                raise
            _remove_busy_dir(path)
    if not exists(path):
        if not ignore_nonexist:
            raise TargetNotExistsError(path)
        return
    os.remove(path)


def move(src, dst, overwrite=True):
    """Move src to dst. If overwrite is False and dst exists,
    a TargetExistsError raises"""
    if not islink(dst) and samefile(src, dst):
        remove(src)
        return
    if overwrite:
        remove(dst)
    elif exists(dst):
        raise TargetExistsError(dst)
    shutil.move(src, dst)


def copy(src, dst, overwrite=True):
    """Copy src to dst. If overwrite is False and dst exists,
    a TargetExistsError raises"""
    if not islink(dst) and samefile(src, dst):
        return
    if overwrite:
        remove(dst)
    elif exists(dst):
        raise TargetExistsError(dst)
    if isdir(src):
        shutil.copytree(src, dst)
    else:
        shutil.copy2(src, dst)


def link(src, dst, overwrite=True):
    """Symbolically link src to dst.
    If overwrite is False and dst exists, a TargetExistsError raises"""
    if not islink(dst) and samefile(src, dst):
        move(dst, src)
        link(src, dst)
        return
    if overwrite:
        remove(dst)
    elif exists(dst):
        raise TargetExistsError(dst)
    os.symlink(src, dst)


def samefile(path1, path2):
    """Tell if two paths are pointing to the same file
    If both of them don't exists, they have to be the same string.
    Otherwise, they have to point to the same file
    """
    exist1 = exists(path1)
    exist2 = exists(path2)
    if not exist1 and not exist2:
        return path1 == path2
    if not exist1 or not exist2:
        return False
    return os.path.samefile(path1, path2)


def makedirs(path, overwrite=True):
    """Create directory. If overwrite is False and path exists,
    a TargetExistsError raises.
    Important: if overwrite is True and path exists,
    content of the directory will lost!"""
    if overwrite:
        remove(path)
    elif exists(path):
        raise TargetExistsError(path)
    os.makedirs(path)


mkdir = makedirs  # pylint: disable=invalid-name


def gzip(src, dst, overwrite=True):
    """Gzip a file or a directory (using tar zip)"""
    if overwrite:
        remove(dst)
    elif exists(dst):
        raise TargetExistsError(dst)
    if os.path.isdir(src):  # tar.gz
        import tarfile
        from glob import glob
        cwd = os.getcwd()
        os.chdir(src)
        with tarfile.open(dst, 'w:gz') as tar:
            for name in glob('./*'):
                tar.add(name)
        os.chdir(cwd)
    else:
        import gzip as _gzip
        with open(src, 'rb') as srcf, _gzip.open(dst, 'wb') as dstf:
            shutil.copyfileobj(srcf, dstf)


def gunzip(src, dst, overwrite=True):
    """Gunzip a gzip file or a tar.gz file"""
    if overwrite:
        remove(dst)
    elif exists(dst):
        raise TargetExistsError(dst)
    if str(src).endswith('.tgz') or str(src).endswith('.tar.gz'):
        import tarfile
        makedirs(dst, overwrite)
        with tarfile.open(src, 'r:gz') as tar:
            tar.extractall(dst)
    else:
        import gzip as _gzip
        with _gzip.open(src, 'rb') as srcf, open(dst, 'wb') as dstf:
            shutil.copyfileobj(srcf, dstf)



"""Implementations for Job"""
from pathlib import Path
from diot import Diot, OrderedDiot
from .logger import logger
from .runner import runnermgr
from .utils import chmod_x, fs, filesig
from .exception import JobInputParseError, JobOutputParseError
from ._proc import (OUT_VARTYPE, OUT_FILETYPE, OUT_DIRTYPE, OUT_STDOUTTYPE,
                    OUT_STDERRTYPE, IN_VARTYPE, IN_FILETYPE, IN_FILESTYPE)

# pylint: disable=unused-argument

RC_NO_RCFILE = 511


def _link_infile(orgfile, indir, basename=None):
    """
	Create links for input files
	@params:
		orgfile (Path): The original input file
        basename (str): use this basename if provided,
                        otherwise use the basename of orgfile
	@returns:
		The link to the original file.
	"""
    infile = indir.joinpath(basename or orgfile.name)
    try:
        fs.link(orgfile, infile, overwrite=False)
    except OSError:
        pass

    if fs.samefile(infile, orgfile):
        return infile

    exist_infiles = (indir).glob('[[]*[]]*')
    num = 0
    for i, eifile in enumerate(exist_infiles):
        # The GIL here takes time, if there are more than 100 files,
        # Just don't check
        if i < 100 and fs.samefile(eifile, orgfile):
            return eifile
        try:
            nexist = int(eifile.name[1:eifile.name.find(']')])
        except ValueError:
            pass
        else:
            num = max(num, nexist)

    infile = indir / '[{}]{}'.format(num + 1, basename or orgfile.name)
    fs.link(orgfile, infile)
    return infile


def job_dir(this, value):
    """Try to create job dir when access it"""
    ret = this.proc.workdir.joinpath(str(this.index + 1)).resolve()
    if not ret.is_dir():
        ret.mkdir(exist_ok=True)
    return ret


def job_input(this, value):  # pylint: disable=too-many-branches
    """
	Prepare input, create link to input files and set other placeholders
	"""
    #fs.remove(this.dir / 'input')
    # fs.makedirs will clean existing dir
    fs.makedirs(this.dir / 'input')
    ret = OrderedDiot()
    for key, val in this.proc.input.items():
        # the original input file(s)
        key = key.strip()
        intype, indata = val[0], val[1][this.index]
        if intype in IN_FILETYPE:
            if not isinstance(indata, (Path, str)):
                raise JobInputParseError(
                    'Not a string or path for input [%s:%s]: %r' %
                    (key, intype, indata))
            if not indata:  # allow empty input
                infile = ''
            elif not fs.exists(indata):
                raise JobInputParseError(
                    'File not exists for input [%s:%s]: %r' %
                    (key, intype, indata))
            else:
                indata = Path(indata)
                infile = _link_infile(indata.resolve(),
                                      this.dir / 'input',
                                      indata.name)
                if indata.name != infile.name:
                    this.logger("Input file renamed: %s -> %s" %
                                (indata.name, infile.name),
                                slevel='INFILE_RENAMING',
                                level="warning")
                indata = str(infile)
            ret[key] = (intype, str(infile))

        elif intype in IN_FILESTYPE:
            ret[key] = (intype, [])

            if not indata:
                this.logger(
                    'No data provided for [%s:%s], use empty list instead.' %
                    (key, intype),
                    slevel='INFILE_EMPTY',
                    level="warning")
                this.data.i[key] = []
                continue

            if not isinstance(indata, list):
                raise JobInputParseError('Not a list for input [%s:%s]: %r' %
                                         (key, intype, indata))

            for i, data in enumerate(indata):
                if not isinstance(data, (Path, str)):
                    raise JobInputParseError(
                        'Not a string or path as an element '
                        'of input [%s:%s]: %r' % (key, intype, data))
                indata[i] = str(data)

                if not data:
                    infile = ''
                elif not fs.exists(data):
                    raise JobInputParseError(
                        'File not exists as an element of input [%s:%s]: %r' %
                        (key, intype, data))
                else:
                    data = Path(data)
                    infile = _link_infile(data.resolve(),
                                          this.dir / 'input',
                                          data.name)
                    if data.name != infile.name:
                        this.logger('Input file renamed: %s -> %s' %
                                    (data.name, infile.name),
                                    slevel='INFILE_RENAMING',
                                    level="warning")
                    indata[i] = str(infile)
                ret[key][1].append(str(infile))
        else:
            ret[key] = (intype, indata)
        this.data.i[key] = indata
    return ret


def job_output(this, value):
    """Build the output data"""
    # keep the output dir if exists
    if not fs.exists(this.dir / 'output'):
        fs.makedirs(this.dir / 'output')

    output = this.proc.output
    # has to be OrderedDict
    assert isinstance(output, dict)
    ret = OrderedDiot()
    # allow empty output
    if not output:
        return ret

    for key, val in output.items():
        outtype, outtpl = val
        outdata = outtpl.render(this.data)

        #this.output[key] = {'type': outtype, 'data': outdata}
        if (outtype in OUT_FILETYPE or
                outtype in OUT_DIRTYPE or
                outtype in OUT_STDOUTTYPE or
                outtype in OUT_STDERRTYPE):
            if Path(outdata).is_absolute():
                raise JobOutputParseError(
                    'Absolute path not allowed for [%s:%s]: %r' %
                    (key, outtype, outdata))
            ret[key] = (outtype, this.dir / 'output' / outdata)
        else:
            ret[key] = (outtype, outdata)
        this.data.o[key] = str(this.dir / 'output' / outdata)
    return ret


def job_signature(this, value):
    """@API
	Calculate the signature of the job based on the input/output and the script.
	If file does not exist, it will not be in the signature.
	The signature is like:
	```json
	{
		"i": {
			"invar:var": <invar>,
			"infile:file": <infile>,
			"infiles:files": [<infiles...>]
		},
		"o": {
			"outvar:var": <outvar>,
			"outfile:file": <outfile>,
			"outdir:dir": <outdir>
		}
		"mtime": <max mtime of input and script>,
	}
	```
	@returns:
		(Diot): The signature of the job
	"""
    script_sig = filesig(this.dir / 'job.script')
    ret = Diot(i={}, o={}, mtime=script_sig[1] if script_sig else 0)

    for key, val in this.input.items():
        (datatype, data) = val
        if datatype in IN_FILETYPE:
            sig = filesig(data, dirsig=this.proc.dirsig)
            if not sig:
                continue
            ret.mtime = max(ret.mtime, sig[1])
            ret.i[key + ":" + IN_FILETYPE[0]] = sig[0]
        elif datatype in IN_FILESTYPE:
            ret.i[key + ":" + IN_FILESTYPE[0]] = []
            for infile in sorted(data):
                sig = filesig(infile, dirsig=this.proc.dirsig)
                if not sig:
                    continue
                ret.mtime = max(ret.mtime, sig[1])
                ret.i[key + ":" + IN_FILESTYPE[0]].append(sig[0])
        else:
            ret.i[key + ":" + IN_VARTYPE[0]] = str(data)

    # flush stat?
    for key, val in this.output.items():
        (datatype, data) = val
        if datatype in OUT_VARTYPE:
            ret.o[key + ":" + OUT_VARTYPE[0]] = str(data)
            continue
        if not fs.exists(data):
            continue
        key += ':'
        key += OUT_FILETYPE[0] if datatype in OUT_FILETYPE else OUT_DIRTYPE[0]
        ret.o[key] = str(data)
    return ret


def job_rc_setter(this, value):
    """Try to save the returncode while set it to job.rc"""
    if value is not None:
        with (this.dir / 'job.rc').open('w') as frc:
            frc.write(str(value))


def job_rc_getter(this, value):
    """Try to read the returncode from job.rc file"""
    rcfile = this.dir / 'job.rc'
    if not fs.isfile(rcfile):
        return RC_NO_RCFILE
    with rcfile.open('r') as frc:
        returncode = frc.read().strip()
        if not returncode or returncode == 'None':
            return RC_NO_RCFILE
        return int(returncode)


def job_data(this, value):
    """
	Data for rendering templates
	@returns:
		(dict): The data used to render the templates.
	"""
    return Diot(job=dict(index=this.index,
                         indir=str(this.dir / 'input'),
                         outdir=str(this.dir / 'output'),
                         dir=str(this.dir),
                         outfile=str(this.dir / 'job.stdout'),
                         errfile=str(this.dir / 'job.stderr'),
                         pidfile=str(this.dir / 'job.pid'),
                         cachedir=str(this.dir / 'output/.jobcache')),
                i={},
                o={},
                proc=this.proc,
                args=this.proc.args)


def job_script(this, value):
    """Try to create the script while accessing it"""
    scriptfile = this.dir.joinpath('job.script.' + this.runner)
    this.logger('Wrapping up script: %s' % scriptfile, level='debug')
    script_parts = this.script_parts

    # redirect stdout and stderr
    if script_parts.saveoe:
        redirect_oe = ''' \\
        1> "$jobdir/job.stdout" \\
        2> "$jobdir/job.stderr"
        '''
        if isinstance(script_parts.command, list):
            script_parts.command[-1] += redirect_oe
        else:
            script_parts.command += redirect_oe

    src = ['#!/usr/bin/env bash']
    srcappend = src.append
    srcextend = src.extend
    addsrc = lambda code: (None
                           if code is None
                           else [srcappend, srcextend][
                               int(isinstance(code, list))
                           ](code))

    addsrc(script_parts.header)
    addsrc('#')
    addsrc('jobdir=%r' % str(this.dir))
    addsrc('# Collect return code on exit')

    trapcmd = ['status=\\$?; \\']
    trapcmd.append('      echo \\$status > \\"$jobdir/job.rc\\"; \\')
    # make sure stdout/err has been created when script exits.
    trapcmd.append('      if [ ! -e \\"$jobdir/job.stdout\\" ]; then \\')
    trapcmd.append('        touch \\"$jobdir/job.stdout\\"; \\')
    trapcmd.append('      fi; \\')
    trapcmd.append('      if [ ! -e \\"$jobdir/job.stderr\\" ]; then \\')
    trapcmd.append('        touch \\"$jobdir/job.stderr\\"; \\')
    trapcmd.append('      fi; \\')
    trapcmd.append('      exit \\$status')
    addsrc('trap "%s" '
           '1 2 3 6 7 8 9 10 11 12 15 16 17 EXIT' % '\n'.join(trapcmd))
    addsrc('')
    addsrc('# START pre-script')
    addsrc(script_parts.pre)
    addsrc('# END pre-script')
    addsrc('')
    addsrc('# START real-script')
    addsrc(script_parts.command)  # pylint: disable=no-member
    addsrc('# END real-script')
    addsrc('')
    addsrc('# START post-script')
    addsrc(script_parts.post)
    addsrc('# END post-script')
    addsrc('')

    scriptfile.write_text('\n'.join(src))
    return chmod_x(scriptfile)


def job_script_parts(this, value):
    """Allow runners to modify the script parts"""
    script = this.proc.script.render(this.data)
    # real script file
    realsfile = this.dir / 'job.script'
    if fs.isfile(realsfile) and realsfile.read_text() != script:
        fs.move(realsfile, this.dir / 'job.script._bak')
        this.logger('Script updated: ' + str(realsfile),
                    slevel='SCRIPT_UPDATED',
                    level='debug')
        realsfile.write_text(script)
    elif not fs.isfile(realsfile):
        realsfile.write_text(script)

    base = Diot(header='',
                pre=this.proc.runner.get(this.runner + '_prescript', ''),
                post=this.proc.runner.get(this.runner + '_postscript', ''),
                saveoe=True,
                command=chmod_x(realsfile))
    ret = runnermgr.hook.script_parts(job=this, base=base)
    return ret or base


def job_logger(this, value):
    """@API
	A logger wrapper to avoid instanize a logger object for each job
	@params:
		*args (str): messages to be logged.
		*kwargs: Other parameters for the logger.
	"""
    def _logger(*args, **kwargs):
        level = kwargs.pop('level', 'info')
        kwargs['proc'] = this.proc.id
        kwargs['jobidx'] = this.index
        kwargs['joblen'] = this.proc.size
        if kwargs.pop('pbar', False):
            logger.pbar[level](*args, **kwargs)
        else:
            logger[level](*args, **kwargs)

    return _logger


def job_pid_setter(this, value):
    """Try to save the pid while setting it to job.pid"""
    if value is None:
        return ''
    this.dir.joinpath('job.pid').write_text(str(value))
    return str(value)


def job_pid_getter(this, value):
    """Try to read the pid from job.pid"""
    if not this.dir.joinpath('job.pid').is_file():
        return ''
    return this.dir.joinpath('job.pid').read_text().strip()



"""
Template adaptor for PyPPL

@variables:
	DEFAULT_ENVS (dict): The default environments for templates
"""
from liquid import Liquid

__all__ = ['Template', 'TemplateLiquid', 'TemplateJinja2']

DEFAULT_ENVS = {}


class Template:
    """@API
	Base class wrapper to wrap template for PyPPL
	"""
    def __init__(self, source, **envs):
        """@API
		Template construct
		"""
        self.source = source
        self.envs = DEFAULT_ENVS.copy()
        self.envs.update(envs)

    def register_envs(self, **envs):
        """@API
		Register extra environment
		@params:
			**envs: The environment
		"""
        self.envs.update(envs)

    def render(self, data=None):
        """@API
		Render the template
		@parmas:
			data (dict): The data used to render
		"""
        return self._render(data or {})

    # in order to dump setting
    def __str__(self):
        lines = self.source.splitlines()
        if len(lines) <= 1:
            return '%s < %s >' % (self.__class__.__name__, ''.join(lines))

        ret = ['%s <<<' % self.__class__.__name__]
        ret += ['\t' + line for line in self.source.splitlines()]
        ret += ['>>>']
        return '\n'.join(ret)

    def __repr__(self):
        return str(self)

    def _render(self, data):
        raise NotImplementedError()


class TemplateLiquid(Template):
    """@API
	liquidpy template wrapper.
	"""
    def __init__(self, source, **envs):
        """
		Initiate the engine with source and envs
		@params:
			`source`: The souce text
			`envs`: The env data
		"""
        super(TemplateLiquid, self).__init__(source, **envs)
        self.envs['__engine'] = 'liquid'
        self.engine = Liquid(source, **self.envs)
        self.source = source

    def _render(self, data):
        """
		Render the template
		@params:
			`data`: The data used for rendering
		@returns:
			The rendered string
		"""
        return self.engine.render(**data)


class TemplateJinja2(Template):
    """@API
	Jinja2 template wrapper
	"""
    def __init__(self, source, **envs):
        """
		Initiate the engine with source and envs
		@params:
			`source`: The souce text
			`envs`: The env data
		"""
        import jinja2
        super(TemplateJinja2, self).__init__(source, **envs)
        self.envs['__engine'] = 'jinja2'
        self.engine = jinja2.Template(source)
        self.engine.globals = self.envs
        self.source = source

    def _render(self, data):
        """
		Render the template
		@params:
			`data`: The data used for rendering
		@returns:
			The rendered string
		"""
        return self.engine.render(data)
